import numpy
cimport numpy
import astropy
import h5py
from ..constants.physics import c

cdef class ImageObject:
    cdef public numpy.ndarray image, x, y, unc, velocity, freq, wave
    cdef public header, wcs

    def __init__(self, numpy.ndarray[double, ndim=4] image=None, \
            numpy.ndarray[double, ndim=1] x=None, \
            numpy.ndarray[double, ndim=1] y=None, \
            numpy.ndarray[double, ndim=1] wave=None, \
            numpy.ndarray[double, ndim=1] freq=None, \
            numpy.ndarray[double, ndim=4] unc=None, \
            numpy.ndarray[double, ndim=1] velocity=None, \
            header=None, wcs = None):

        self.image = image

        self.x = x
        self.y = y

        self.header = header
        self.wcs = wcs

        self.unc = unc
        self.velocity = velocity

        if (type(wave) == type(None)) and (type(freq) != type(None)):
            self.freq = freq
            self.wave = c / freq
        elif (type(wave) != type(None)) and (type(freq) == type(None)):
            self.wave = wave
            self.freq = c / wave
        elif (type(wave) != type(None)) and (type(freq) != type(None)):
            self.wave = wave
            self.freq = freq

    def __reduce__(self):
        return (rebuild, (self.image, self.x, self.y, self.wave, self.freq, \
                self.unc, self.velocity, self.header, self.wcs))

def rebuild(image, x, y, wave, freq, unc, velocity, header, wcs):
    return ImageObject(image, x, y, wave, freq, unc, velocity, header, wcs)

class Image(ImageObject):

    def asFITS(self, channel=None):
        hdulist = astropy.io.fits.HDUList([])

        if channel != None:
            hdu = astropy.io.fits.PrimaryHDU(self.image[:,:,channel,0].astype(\
                    numpy.float32))

            if type(self.header) != type(None):
                hdu.header = self.header[channel]

            hdulist.append(hdu)

        else:
            hdu = astropy.io.fits.PrimaryHDU(numpy.transpose(\
                    self.image[:,:,:,0].astype(numpy.float32), axes=[2,1,0]))

            if type(self.header) != type(None):
                hdu.header = self.header

            hdulist.append(hdu)

        return hdulist

    def set_uncertainty_from_image(self, image, box=128):
        cdef unsigned int ny, nx, nfreq, npol
        cdef unsigned int xmin, xmax, ymin, ymax
        cdef unsigned int i, j
        cdef unsigned int halfbox
        cdef numpy.ndarray[double, ndim=2] subimage
        cdef numpy.ndarray[double, ndim=4] unc

        ny, nx, nfreq, npol = self.image.shape

        halfbox = box / 2

        unc = numpy.empty(self.image.shape)
        for i in range(ny):
            for j in range(nx):
                xmin = max(0,<int>j-64)
                xmax = min(<int>j+64,nx)
                ymin = max(0,<int>i-64)
                ymax = min(<int>i+64,ny)

                subimage = image.image[ymin:ymax,xmin:xmax,0,0]

                if (numpy.isnan(subimage).prod() == True):
                    unc[i,j,0,0] = subimage[0,0]
                else:
                    unc[i,j,0,0] = numpy.nanstd(subimage)

        self.unc = unc

    def read(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "r")
        else:
            f = usefile

        if ('x' in f):
            x = f['x'][...].astype(numpy.double)
            y = f['y'][...].astype(numpy.double)
        else:
            x = None
            y = None

        if ('freq' in f):
            freq = f['freq'][...].astype(numpy.double)
        else:
            freq = None

        image = f['image'][...].astype(numpy.double)

        if ('unc' in f):
            unc = f['unc'][...].astype(numpy.double)
        else:
            unc = None

        if ('wcs' in f.attrs):
            wcs = astropy.wcs.WCS(f.attrs['wcs'])
        else:
            wcs = None

        if ('header' in f.attrs):
            header = astropy.io.fits.Header().fromstring(f.attrs['header'])
        else:
            header = None

        self.__init__(image, x=x, y=y, unc=unc, freq=freq, wcs=wcs, \
                header=header)

        if (usefile == None):
            f.close()

    def write(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "w")
        else:
            f = usefile

        #if hasattr(self, "x"):
        if (type(self.x) != type(None)):
            x_dset = f.create_dataset("x", self.x.shape, dtype='f')
            x_dset[...] = self.x
            y_dset = f.create_dataset("y", self.y.shape, dtype='f')
            y_dset[...] = self.y

        #if hasattr(self, "freq"):
        if (type(self.freq) != type(None)):
            freq_dset = f.create_dataset("freq", self.freq.shape, dtype='f')
            freq_dset[...] = self.freq

        image_dset = f.create_dataset("image", self.image.shape, dtype='f')
        image_dset[...] = self.image

        #if hasattr(self, "unc"):
        if (type(self.unc) != type(None)):
            unc_dset = f.create_dataset("unc", self.unc.shape, \
                    dtype='f')
            unc_dset[...] = self.unc

        if hasattr(self, "wcs") and ((type(self.wcs) != type(None))):
            f.attrs['wcs'] = self.wcs.to_header_string()

        if hasattr(self, "header") and (type(self.header) != type(None)):
            f.attrs['header'] = self.header.tostring()

        if (usefile == None):
            f.close()

################################################################################
#
# Unstructured images.
#
################################################################################


cdef class UnstructuredImageObject:
    cdef public numpy.ndarray image, x, y, unc, velocity, freq, wave

    def __init__(self, numpy.ndarray[double, ndim=2] image=None, \
            numpy.ndarray[double, ndim=1] x=None, \
            numpy.ndarray[double, ndim=1] y=None, \
            numpy.ndarray[double, ndim=1] wave=None, \
            numpy.ndarray[double, ndim=1] freq=None, \
            numpy.ndarray[double, ndim=2] unc=None, \
            numpy.ndarray[double, ndim=1] velocity=None):

        self.image = image

        self.x = x
        self.y = y

        self.unc = unc
        self.velocity = velocity

        if (type(wave) == type(None)) and (type(freq) != type(None)):
            self.freq = freq
            self.wave = c / freq
        elif (type(wave) != type(None)) and (type(freq) == type(None)):
            self.wave = wave
            self.freq = c / wave
        elif (type(wave) != type(None)) and (type(freq) != type(None)):
            self.wave = wave
            self.freq = freq

    def __reduce__(self):
        return (rebuild_unstructured, (self.image, self.x, self.y, self.wave, \
                self.freq, self.unc, self.velocity))

def rebuild_unstructured(image, x, y, wave, freq, unc, velocity):
    return UnstructuredImageObject(image, x, y, wave, freq, unc, velocity)

class UnstructuredImage(UnstructuredImageObject):

    def read(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "r")
        else:
            f = usefile

        if ('x' in f):
            x = f['x'][...].astype(numpy.double)
            y = f['y'][...].astype(numpy.double)
        else:
            x = None
            y = None

        if ('freq' in f):
            freq = f['freq'][...].astype(numpy.double)
        else:
            freq = None

        image = f['image'][...].astype(numpy.double)

        if ('unc' in f):
            unc = f['unc'][...].astype(numpy.double)
        else:
            unc = None

        self.__init__(image, x=x, y=y, unc=unc, freq=freq)

        if (usefile == None):
            f.close()

    def write(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "w")
        else:
            f = usefile

        if (type(self.x) != type(None)):
            x_dset = f.create_dataset("x", self.x.shape, dtype='f')
            x_dset[...] = self.x
            y_dset = f.create_dataset("y", self.y.shape, dtype='f')
            y_dset[...] = self.y

        if (type(self.freq) != type(None)):
            freq_dset = f.create_dataset("freq", self.freq.shape, dtype='f')
            freq_dset[...] = self.freq

        image_dset = f.create_dataset("image", self.image.shape, dtype='f')
        image_dset[...] = self.image

        if (type(self.unc) != type(None)):
            unc_dset = f.create_dataset("unc", self.unc.shape, \
                    dtype='f')
            unc_dset[...] = self.unc

        if (usefile == None):
            f.close()



import numpy
import scipy
from .Dust import Dust
from ..constants.math import pi

def mix_dust(dust, abundance, medium=None, rule="Bruggeman", filling=1.):

    if rule == "Bruggeman":
        meff = numpy.zeros(dust[0].lam.size,dtype=complex)
        rho = 0.0
        
        for i in range(dust[0].lam.size):
            temp = scipy.optimize.fsolve(bruggeman,numpy.array([1.0,0.0]),\
                    args=(dust,abundance,i, filling))
            meff[i] = temp[0]+1j*temp[1]
        
        for i in range(len(dust)):
            rho += dust[i].rho*abundance[i]

        rho *= filling
    
    elif rule == "MaxGarn":
        numerator = 0.0+1j*0.0
        denominator = 0.0+1j*0.0
        rho = 0.0
        
        for i in range(len(dust)):
            gamma = 3. / (dust[i].m**2 + 2)

            numerator += abundance[i] * gamma * dust[i].m**2
            denominator += abundance[i] * gamma

            rho += dust[i].rho*abundance[i]

        mmix = numpy.sqrt(numerator / denominator)
        
        F = (mmix**2 - 1.) / (mmix**2 + 2.)

        meff = numpy.sqrt((1. + 2.*filling*F) / (1. - filling*F))

        rho *= filling

    new = Dust()
    new.set_density(rho)
    new.set_optical_constants(dust[0].lam, meff.real, meff.imag)
    
    return new

def bruggeman(meff, dust, abundance, index, filling):
    
    m_eff = meff[0]+1j*meff[1]
    tot = 0+0j
    
    for j in range(len(dust)):
        tot += filling * abundance[j]*(dust[j].m[index]**2-m_eff**2)/ \
                (dust[j].m[index]**2+2*m_eff**2)

    # Add in the void.

    tot += (1 - filling) * (1. - m_eff**2) / (1. + 2*m_eff**2)
    
    return numpy.array([tot.real,tot.imag])



from os import system

import sys
if sys.version_info.major > 2:
    from subprocess import STDOUT, run
else:
    from subprocess32 import STDOUT, run

def thermal(noscat=None, nphot_therm=None, nphot_scat=None, setthreads=1, \
        inclfreefree=None, nofreefree=None, inclgascont=None, nogascont=None, \
        verbose=True, timelimit=7200, nice=None):

    if nice != None:
        command="nice -{0:d} radmc3d mctherm".format(nice)
    else:
        command="radmc3d mctherm"

    if (noscat == True):
        command += " noscat"
    if (nphot_therm != None):
        command += " nphot_therm {0:d}".format(nphot_therm)
    if (setthreads != 1):
        command += " setthreads {0:d}".format(setthreads)
    if (inclfreefree == True):
        command += " inclfreefree"
    if (nofreefree == True):
        command += " nofreefree"
    if (inclgascont == True):
        command += " inclgascont"
    if (nogascont == True):
        command += " nogascont"

    if not verbose:
        f = open("radmc3d.out","w")
        output = run(command.split(" "), stdout=f, stderr=f, timeout=timelimit)
        f.close()
    else:
        output = run(command.split(" "), stderr=STDOUT, timeout=timelimit)

def scattering(nphot_scat=None, setthreads=1, inclfreefree=None, \
        nofreefree=None, inclgascont=None, nogascont=None, \
        verbose=True, nice=None, loadlambda=None):

    if nice != None:
        command="nice -{0:d} radmc3d mcscat".format(nice)
    else:
        command="radmc3d mcscat "

    if (nphot_scat != None):
        command += " nphot_scat {0:d}".format(nphot_therm)
    if (setthreads != 1):
        command += " setthreads {0:d}".format(setthreads)
    if (inclfreefree == True):
        command += " inclfreefree"
    if (nofreefree == True):
        command += " nofreefree"
    if (inclgascont == True):
        command += " inclgascont"
    if (nogascont == True):
        command += " nogascont"
    if (loadlambda == True):
        command += "loadlambda "

    if not verbose:
        command += " > radmc3d.out"

    system(command)

def sed(nrrefine=None, fluxcons=None, norefine=None, nofluxcons=None, \
        noscat=None, sizeau=None, sizepc=None, zoomau=None, zoompc=None, \
        truepix=None, truezoom=None, pointau=None, pointpc=None, incl=None, \
        phi=None, posang=None, circ=None, apert=None, useapert=None, \
        noapert=None, nphot_scat=None, inclstar=None, nostar=None, \
        inclline=None, noline=None, incldust=None, nodust=None, \
        inclfreefree=None, nofreefree=None, inclgascont=None, nogascont=None, \
        loadlambda=None, verbose=True, nice=None):

    if nice != None:
        command="nice -{0:d} radmc3d spectrum ".format(nice)
    else:
        command="radmc3d spectrum "

    if (nrrefine != None):
        command += "nrrefine {0:i} ".format(nrrefine)
    if (fluxcons == True):
        command += "fluxcons "
    if (norefine == True):
        command += "norefine "
    if (nofluxcons == True):
        command += "nofluxcons "
    if (noscat == True):
        command += "noscat "
    if (sizeau != None):
        command += "sizeau {0:f} ".format(sizeau)
    if (sizepc != None):
        command += "sizepc {0:f} ".format(sizepc)
    if (zoomau != None):
        command += "zoomau {0:f} {1:f} {2:f} {3:f} ".format(zoomau[0], \
                zoomau[1], zoomau[2], zoomau[3])
    if (zoompc != None):
        command += "zoompc {0:f} {1:f} {2:f} {3:f} ".format(zoompc[0], \
                zoompc[1], zoompc[2], zoompc[3])
    if (truepix == True):
        command += "truepix "
    if (truezoom == True):
        command += "truezoom "
    if (pointau != None):
        command += "pointau {0:f} {1:f} {2:f} ".format(pointau[0], pointau[1], \
                pointau[2])
    if (pointpc != None):
        command += "pointpc {0:f} {1:f} {2:f} ".format(pointpc[0], pointpc[1], \
                pointpc[2])
    if (incl != None):
        command += "incl {0:f} ".format(incl)
    if (phi != None):
        command += "phi {0:f} ".format(phi)
    if (posang != None):
        command += "posang {0:f} ".format(posang)
    if (circ == True):
        command += "circ "
    if (apert == True):
        command += "apert "
    if (useapert == True):
        command += "useapert "
    if (noapert == True):
        command += "noapert "
    if (nphot_scat != None):
        command += "nphot_scat {0:d} ".format(nphot_scat)
    if (inclstar == True):
        command += "inclstar "
    if (nostar == True):
        command += "nostar "
    if (inclline == True):
        command += "inclline "
    if (noline == True):
        command += "noline "
    if (incldust == True):
        command += "incldust "
    if (nodust == True):
        command += "nodust "
    if (inclfreefree == True):
        command += "inclfreefree "
    if (nofreefree == True):
        command += "nofreefree "
    if (inclgascont == True):
        command += "inclgascont "
    if (nogascont == True):
        command += "nogascont "
    if (loadlambda == True):
        command += "loadlambda "

    if not verbose:
        command += " > radmc3d.out"

    system(command)

def image(lam=None, npix=None, npixx=None, npixy=None, nrrefine=None, \
        fluxcons=None, norefine=None, nofluxcons=None, noscat=None, \
        ilambda=None, inu=None, color=None, loadcolor=None, loadlambda=None, \
        sizeau=None, sizepc=None, zoomau=None, zoompc=None, truepix=None, \
        truezoom=None, pointau=None, pointpc=None, incl=None, phi=None, \
        posang=None, imageunform=None, imageformatted=None, tracetau=None, \
        tracecolumn=None, tracenormal=None, apert=None, useapert=None, \
        noapert=None, nphot_scat=None, inclstar=None, nostar=None, \
        inclline=None, noline=None, incldust=None, nodust=None, \
        inclfreefree=None, nofreefree=None, inclgascont=None, nogascont=None, \
        widthkms=None, vkms=None, linenlam=None, iline=None, imolspec=None, \
        doppcatch=None, verbose=True, nice=None, unstructured=False, \
        circ=False):

    if nice != None:
        command="nice -{0:d} radmc3d image ".format(nice)
    else:
        command="radmc3d image "

    if (circ):
        command += "circ "
    if (lam != None):
        command += "lambda "+lam+" "
    if (iline != None):
        command += "iline {0:d} ".format(iline)
    if (imolspec != None):
        command += "imolspec {0:d} ".format(imolspec)
    if (widthkms != None):
        command += "widthkms {0:f} ".format(widthkms)
    if (vkms != None):
        command += "vkms {0:f} ".format(vkms)
    if (doppcatch == True):
        command += "doppcatch "
    if (linenlam != None):
        command += "linenlam {0:d} ".format(linenlam)
    if (npix != None):
        command += "npix {0:d} ".format(npix)
    if (npixx != None):
        command += "npixx {0:i} ".format(npixx)
    if (npixy != None):
        command += "npixy {0:i} ".format(npixy)
    if (nrrefine != None):
        command += "nrrefine {0:i} ".format(nrrefine)
    if (fluxcons == True):
        command += "fluxcons "
    if (norefine == True):
        command += "norefine "
    if (nofluxcons == True):
        command += "nofluxcons "
    if (noscat == True):
        command += "noscat "
    if (ilambda != None):
        command += "ilambda {0:i} ".format(ilambda)
    if (inu != None):
        command += "inu {0:i} ".format(inu)
    if (color == True):
        command += "color "
    if (loadcolor == True):
        command += "loadcolor "
    if (loadlambda == True):
        command += "loadlambda "
    if (sizeau != None):
        command += "sizeau {0:f} ".format(sizeau)
    if (sizepc != None):
        command += "sizepc {0:f} ".format(sizepc)
    if (zoomau != None):
        command += "zoomau {0:f} {1:f} {2:f} {3:f} ".format(zoomau[0], \
                zoomau[1], zoomau[2], zoomau[3])
    if (zoompc != None):
        command += "zoompc {0:f} {1:f} {2:f} {3:f} ".format(zoompc[0], \
                zoompc[1], zoompc[2], zoompc[3])
    if (truepix == True):
        command += "truepix "
    if (truezoom == True):
        command += "truezoom "
    if (pointau != None):
        command += "pointau {0:f} {1:f} {2:f} ".format(pointau[0], pointau[1], \
                pointau[2])
    if (pointpc != None):
        command += "pointpc {0:f} {1:f} {2:f} ".format(pointpc[0], pointpc[1], \
                pointpc[2])
    if (incl != None):
        command += "incl {0:f} ".format(incl)
    if (phi != None):
        command += "phi {0:f} ".format(phi)
    if (posang != None):
        command += "posang {0:f} ".format(posang)
    if (imageunform == True):
        command += "imageunform "
    if (imageformatted == True):
        command += "imageformatted "
    if (tracetau == True):
        command += "tracetau "
    if (tracecolumn == True):
        command += "tracecolumn "
    if (tracenormal == True):
        command += "tracenormal "
    if (apert == True):
        command += "apert "
    if (useapert == True):
        command += "useapert "
    if (noapert == True):
        command += "noapert "
    if (nphot_scat != None):
        command += "nphot_scat {0:i} ".format(nphot_scat)
    if (inclstar == True):
        command += "inclstar "
    if (nostar == True):
        command += "nostar "
    if (inclline == True):
        command += "inclline "
    if (noline == True):
        command += "noline "
    if (incldust == True):
        command += "incldust "
    if (nodust == True):
        command += "nodust "
    if (inclfreefree == True):
        command += "inclfreefree "
    if (nofreefree == True):
        command += "nofreefree "
    if (inclgascont == True):
        command += "inclgascont "
    if (nogascont == True):
        command += "nogascont "

    if unstructured:
        command += "diag_subpix "

    if not verbose:
        command += " > radmc3d.out"

    system(command)


# The code below has every single command, if you need to implement more command line options.
"""
def run_thermal(npix=None, npixx=None, npixy=None, nrrefine=None, \
        fluxcons=None, norefine=None, nofluxcons=None, noscat=None, \
        ilambda=None, inu=None, color=None, loadcolor=None, loadlambda=None, \
        sizeau=None, sizepc=None, zoomau=None, zoompc=None, truepix=None, \
        truezoom=None, pointau=None, pointpc=None, incl=None, phi=None, \
        posang=None, imageunform=None, imageformatted=None, circ=None, \
        tracetau=None, tracecolumn=None, tracenormal=None, apert=None, \
        useapert=None, noapert=None, nphot_therm=None, nphot_scat=None, \
        nphot_mcmono=None, inclstar=None, nostar=None, inclline=None, \
        noline=None, incldust=None, nodust=None, inclfreefree=None, \
        nofreefree=None, inclgascont=None, nogascont=None):

    command="radmc3d mctherm "

    if (npix != None):
        command += "npix {0:i} ".format(npix)
    if (npixx != None):
        command += "npixx {0:i} ".format(npixx)
    if (npixy != None):
        command += "npixy {0:i} ".format(npixy)
    if (nrrefine != None):
        command += "nrrefine {0:i} ".format(nrrefine)
    if (fluxcons == True):
        command += "fluxcons "
    if (norefine == True):
        command += "norefine "
    if (nofluxcons == True):
        command += "nofluxcons "
    if (noscat == True):
        command += "noscat "
    if (ilambda != None):
        command += "ilambda {0:i} ".format(ilambda)
    if (inu != None):
        command += "inu {0:i} ".format(inu)
    if (color == True):
        command += "color "
    if (loadcolor == True):
        command += "loadcolor "
    if (loadlambda == True):
        command += "loadlambda "
    if (sizeau != None):
        command += "sizeau {0:f} ".format(sizeau)
    if (sizepc != None):
        command += "sizepc {0:f} ".format(sizepc)
    if (zoomau != None):
        command += "zoomau {0:f} {1:f} {2:f} {3:f} ".format(zoomau[0], \
                zoomau[1], zoomau[2], zoomau[3])
    if (zoompc != None):
        command += "zoompc {0:f} {1:f} {2:f} {3:f} ".format(zoompc[0], \
                zoompc[1], zoompc[2], zoompc[3])
    if (truepix == True):
        command += "truepix "
    if (truezoom == True):
        command += "truezoom "
    if (pointau != None):
        command += "pointau {0:f} {1:f} {2:f} ".format(pointau[0], pointau[1], \
                pointau[2])
    if (pointpc != None):
        command += "pointpc {0:f} {1:f} {2:f} ".format(pointpc[0], pointpc[1], \
                pointpc[2])
    if (incl != None):
        command += "incl {0:f} ".format(incl)
    if (phi != None):
        command += "phi {0:f} ".format(phi)
    if (posang != None):
        command += "posang {0:f} ".format(posang)
    if (imageunform == True):
        command += "imageunform "
    if (imageformatted == True):
        command += "imageformatted "
    if (circ == True):
        command += "circ "
    if (tracetau == True):
        command += "tracetau "
    if (tracecolumn == True):
        command += "tracecolumn "
    if (tracenormal == True):
        command += "tracenormal "
    if (apert == True):
        command += "apert "
    if (useapert == True):
        command += "useapert "
    if (noapert == True):
        command += "noapert "
    if (nphot_therm != None):
        command += "nphot_therm {0:i} ".format(nphot_therm)
    if (nphot_scat != None):
        command += "nphot_scat {0:i} ".format(nphot_scat)
    if (nphot_mcmono != None):
        command += "nphot_mcmono {0:i} ".format(nphot_mcmono)
    if (inclstar == True):
        command += "inclstar "
    if (nostar == True):
        command += "nostar "
    if (inclline == True):
        command += "inclline "
    if (noline == True):
        command += "noline "
    if (incldust == True):
        command += "incldust "
    if (nodust == True):
        command += "nodust "
    if (inclfreefree == True):
        command += "inclfreefree "
    if (nofreefree == True):
        command += "nofreefree "
    if (inclgascont == True):
        command += "inclgascont "
    if (nogascont == True):
        command += "nogascont "

    system(command)
"""



from mpl_toolkits.axes_grid1 import make_axes_locatable
from ..interferometry import Visibilities, clean
import matplotlib.ticker as ticker
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import astropy.stats
import numpy

def plot_continuum_image(visibilities, model, parameters, params, index=0, \
        fig=None, cmap="jet", fontsize="medium", image="data", \
        contours="model", model_image="beam-convolve", \
        weighting="robust", robust=2, maxiter=200, threshold=0.001, 
        uvtaper=None, cmap_contours="none", colors_contours="none",levels=None,\
        negative_levels=None, show_beam=False, beamxy=(0.1,0.1), \
        show_colorbar=False, cax=None, colorbar_location='right', \
        colorbar_orientation='vertical', colorbar_size='5%', colorbar_pad=0.05,\
        units="Jy/beam"):
    r"""
    Plot the millimeter continuum image, along with the best fit model.

    Args:
        :attr:`visibilities` (`dict`):
            Dictionary containing the visibility data, typically as loaded by 
            :code:`utils.load_config` and :code:`utils.load_data`.
        :attr:`model` (`modeling.Model`):
            The radiative transfer model that you would like to plot the 
            visibilities of. Typically this is the output of 
            modeling.run_disk_model.
        :attr:`parameters` (`dict`):
            The parameters dictionary in the config module as loaded in by 
            :code:`utils.load_config`
        :attr:`params` (`dict`):
            The parameters of the model, typically as a dictionary mapping 
            parameter keys from the :code:`parameters` dictionary to their 
            values.
        :attr:`index` (`int`, optional):
            The visibilities dictionary typically contains a list of datasets. 
            `index` indicates which one to plot.
        :attr:`fig` (`tuple`, `(matplotlib.Figure, matplotlib.Axes)`, optional):
            If you've already created a figure and axes to put the plot in, you 
            can supply them here. Otherwise, `plot_1D_visibilities` will 
            generate them for you. Default: `None`
        :attr:`image` (`str`, optional):
            Should the image show the `"data"`, `"model"`, or `"residuals"`.
            Default: `"data"`
        :attr:`contours` (`str`, optional):
            Should the image show the `"data"`, `"model"`, or `"residuals"`.
            Or could also not show contours with `"none"`. Default: `"model"`
        :attr:`model_image` (`str`, optional):
            Should the model image be made by convolving a radiatie transfer
            generated image with an estimate of the beam (`"beam-convolve"`), 
            or by generating model visibilities at the correct baselines of the
            data and then using :code:`interferometry.clean` to generate an 
            image (`"CLEAN"`). Default: `"beam-convolve"`
        :attr:`weighting` (`str`, optional):
            What weighting scheme should the model image use if 
            :code:`model_image="CLEAN"`, `"natural"`, `"robust"`, or 
            `"uniform"`. Default:`"robust"`
        :attr:`robust` (`int`, optional):
            The robust parameter when :code:`weighting="robust"`. Default: 2
        :attr:`maxiter` (`int`, optional):
            The maximum number of CLEAN iterations to perform. Default: 2
        :attr:`threshold` (`float`, optional, Jy):
            The stopping threshold for the CLEAN algorithm. Default: 0.001 Jy
        :attr:`uvtaper` (`float`, optional, lambda):
            FWHM of the Gaussian tapering to apply to the weights when running
            CLEAN. Default: None
        :attr:`cmap` (`str`, optional):
            Which colormap to use for plotting the image. Default: `"jet"`
        :attr:`cmap_contours` (`str`, optional):
            Colormap to use for the contours. If `"none"`, use the same as 
            :code:`cmap`. Default: `"none"`
        :attr:`colors_contours` (`str` or `list`-like, optional):
            Colors to use for the contours. If `"none"`, use the :code:`"cmap"`
            from the image. Default: `"none"`
        :attr:`levels` (`list` of `float`, optional):
            The flux levels at which to plot contours. If `None`, use 
            :code:`[0.1, 0.3, 0.5, 0.7, 0.9] x image.max()`. Default: `None`
        :attr:`negative_levels` (`list` of `float`, optional):
            A list of negative flux levels at which to plot dashed contours.
            Default: `None`
        :attr:`show_beam` (`bool`, optional):
            Should the plot show the size of the beam? Default: `False`
        :attr:`beamxy` (`tuple`, optional):
            If :code:`show_beam=True`, where should the beam be placed in the
            image, in units of axes fraction. Default: `(0.1,0.1)`
        :attr:`show_colorbar` (`bool`, optional):
            Should the plot show a colorbar for the image? Default: `False` 
        :attr:`cax` (`matplotlib.Axes`, optional):
            Pass an existing Axes class to use for the colorbar. If `None`, 
            :code:`plot_continuum_image` will generate a new one. 
            Default:`None`
        :attr:`colorbar_location` (`str`, optional):
            Should the colorbar be located to the `"right"` of the image or 
            on `"top"` of the image. Default: `'right'`
        :attr:`colorbar_size` (`str`, optional):
            The percent width of the image Axes that should be used for the 
            width of the colorbar Axes. Default: '5%'
        :attr:`colorbar_pad` (`str`, optional):
            How much padding should there be between the image Axes and the 
            colorbar Axes. Default: `0.05`
        :attr:`units` (`str`, optional):
            What units should the colorbar use? Options include `"Jy/beam"`, 
            `"mJy/beam"`, and `"uJy/beam"`. Default: `"Jy/beam"`
        :attr:`fontsize` (`str` or `int`):
            What fontsize to use for labels, ticks, etc. Default: `"medium"`

    Returns:
        :attr:`fig` (`matplotlib.Figure`):
            The matplotlib figure that was used for the plot.
        :attr:`ax` (`matplotlib.Axes`):
            The matplotlib axes that were used for the plot.
        :attr:`cax` (`matplotlib.Axes`, optional):
            The matplotlib axes that were used for the colorbar, if 
            :code:`show_colorbar=True`.
    """

    # If no figure is provided, create one.

    if fig == None:
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(4.5,4))
    else:
        fig, ax = fig

    # If no contours colormap is provided, use the same as image.

    if cmap_contours == "none" and colors_contours == "none":
        cmap_contours = cmap
        colors_contours = None
    elif cmap_contours == "none" and colors_contours != "none":
        cmap_contours = None
    elif cmap_contours != "none" and colors_contours == "none":
        colors_contours = None
    elif cmap_contours != "none" and colors_contours != "none":
        colors_contours = None

    # Get the ticks.

    ticks = visibilities["image_ticks"][index]

    # Make the plot.

    for i in range(2):
        # First pass through plot the image, second pass through, plot contours.
        if i == 0:
            plot_type = image
        else:
            plot_type = contours

            # If we don't want to plot contours, just skip the second round.
            if plot_type == "none":
                continue

        # Get the appropriate image (data, model, residual) for plotting.

        if plot_type == "data":
            plot_image = visibilities["image"][index]
        elif plot_type == "model":
            if model_image == "beam-convolve":
                plot_image = model.images[visibilities["lam"][index]]
            else:
                model.visibilities[visibilities["lam"][index]].weights = \
                    visibilities["data"][index].weights

                plot_image = clean(\
                        model.visibilities[visibilities["lam"][index]], \
                        imsize=visibilities["image_npix"][index], \
                        pixel_size=visibilities["image_pixelsize"][index], \
                        weighting=weighting, robust=robust, \
                        convolution="expsinc", mfs=False, mode="continuum", \
                        maxiter=maxiter, threshold=threshold, \
                        uvtaper=uvtaper)[0]
        elif plot_type == "residuals":
            residuals = Visibilities(visibilities["data"][index].u, \
                    visibilities["data"][index].v, \
                    visibilities["data"][index].freq, \
                    visibilities["data"][index].real.copy(), \
                    visibilities["data"][index].imag.copy(),\
                    visibilities["data"][index].weights)

            residuals.real -= model.visibilities[visibilities["lam"]\
                    [index]].real
            residuals.imag -= model.visibilities[visibilities["lam"]\
                    [index]].imag

            plot_image = clean(residuals, \
                    imsize=visibilities["image_npix"][index], \
                    pixel_size=visibilities["image_pixelsize"][index], \
                    weighting=weighting, robust=robust, convolution="expsinc", \
                    mfs=False, mode="continuum", maxiter=maxiter, \
                    uvtaper=uvtaper)[0]

        # Get the appropriate scaling for the image.

        if units == "mJy/beam":
            scale = 1000
            colorbar_label = "mJy beam$^{-1}$"
        elif units == "uJy/beam":
            scale = 1e6
            colorbar_label = "$\mu$Jy beam$^{-1}$"
        else:
            scale = 1.
            colorbar_label = "Jy beam$^{-1}$"

        # Get the contour levels if none are specified.

        if i == 1:
            if levels is None:
                levels = numpy.array([0.1, 0.3, 0.5, 0.7, 0.9])*\
                        plot_image.image.max()*scale

                # Only include levels above some detection threshold.

                rms = astropy.stats.mad_std(visibilities["image"][index].\
                        image[:,:,0,0])*scale * plot_image.image.max() / \
                        visibilities["image"][index].image.max()

                levels = levels[levels > 3*rms]

        # Get the right xmin, xmax, ymin, ymax.

        if plot_type == "data":
            if "x0" in params:
                x0 = -params["x0"]/visibilities["image_pixelsize"][index]
            else:
                x0 = -parameters["x0"]["value"]/\
                        visibilities["image_pixelsize"][index]
            if "y0" in params:
                y0 = params["y0"]/visibilities["image_pixelsize"][index]
            else:
                y0 = parameters["y0"]["value"]/\
                        visibilities["image_pixelsize"][index]

            xmin = int(round(x0 + visibilities["image_npix"][index]/2 - \
                    visibilities["x0"][index]/\
                    visibilities["image_pixelsize"][index]+ \
                    ticks[0]/visibilities["image_pixelsize"][index]))
            xmax = int(round(x0 + visibilities["image_npix"][index]/2 - \
                    visibilities["x0"][index]/\
                    visibilities["image_pixelsize"][index]+ \
                    ticks[-1]/visibilities["image_pixelsize"][index]))

            ymin = int(round(y0 + visibilities["image_npix"][index]/2 + \
                    visibilities["y0"][index]/\
                    visibilities["image_pixelsize"][index]+ \
                    ticks[0]/visibilities["image_pixelsize"][index]))
            ymax = int(round(y0 + visibilities["image_npix"][index]/2 + \
                    visibilities["y0"][index]/\
                    visibilities["image_pixelsize"][index]+ \
                    ticks[-1]/visibilities["image_pixelsize"][index]))
        elif plot_type in ["model","residuals"]:
            xmin, xmax = int(round(visibilities["image_npix"][index]/2+1 + \
                    ticks[0]/visibilities["image_pixelsize"][index])), \
                    int(round(visibilities["image_npix"][index]/2+1 + \
                    ticks[-1]/visibilities["image_pixelsize"][index]))
            ymin, ymax = int(round(visibilities["image_npix"][index]/2+1 + \
                    ticks[0]/visibilities["image_pixelsize"][index])), \
                    int(round(visibilities["image_npix"][index]/2+1 + \
                    ticks[-1]/visibilities["image_pixelsize"][index]))

        if i == 0:
            # Plot the image.

            implot = ax.imshow(plot_image.image[ymin:ymax,xmin:xmax,0,0]*\
                    scale, origin="lower", interpolation="nearest", cmap=cmap)

            if show_beam:
               bmaj = visibilities["image"][index].header["BMAJ"]/\
                       abs(visibilities["image"][index].header["CDELT1"])
               bmin = visibilities["image"][index].header["BMIN"]/\
                       abs(visibilities["image"][index].header["CDELT1"])
               bpa = visibilities["image"][index].header["BPA"]

               xy = ((xmax - xmin)*beamxy[0], (ymax - ymin)*beamxy[1])

               ax.add_artist(patches.Ellipse(xy=xy, width=bmaj, \
                       height=bmin, angle=(bpa+90), facecolor="white", \
                       edgecolor="black"))

            # Add a colorbar to the image.

            if show_colorbar:
                # If no cax were provided, create them based on the options.

                user_specified_cbar = True

                if cax == None:
                    divider = make_axes_locatable(ax)

                    cax = divider.append_axes(colorbar_location, \
                            size=colorbar_size, pad=colorbar_pad)

                    if colorbar_location in ['top','bottom']:
                        colorbar_orientation = "horizontal"
                    else:
                        colorbar_orientation = "vertical"

                    user_specified_cbar = False

                # Now plot the colorbar.

                cbar = plt.colorbar(implot, cax=cax, \
                        orientation=colorbar_orientation)

                # And make some adustments.

                if not user_specified_cbar:
                    if colorbar_location == 'top':
                        cax.xaxis.set_ticks_position('top')
                        cax.xaxis.set_label_position('top')
                    elif colorbar_location == 'left':
                        cax.yaxis.set_ticks_position('left')
                        cax.yaxis.set_label_position('left')

                cbar.set_label(colorbar_label, size=fontsize)

                cax.tick_params(axis="both", which="major", labelsize=fontsize)
        else:
            # Contour the model over the data.

            if len(numpy.where(plot_image.image[ymin:ymax,xmin:xmax,0,0]*\
                    scale > levels.min())[0]) > 0:
                ax.contour(plot_image.image[ymin:ymax,xmin:xmax,0,0]*scale, \
                        cmap=cmap_contours, colors=colors_contours, \
                        levels=levels)

            if negative_levels is not None:
                if len(numpy.where(plot_image.image[ymin:ymax,xmin:xmax,0,0]*\
                        scale < negative_levels.max())[0]) > 0:
                    ax.contour(plot_image.image[ymin:ymax,xmin:xmax,0,0]*scale,\
                            cmap=cmap_contours, colors=colors_contours, \
                            levels=negative_levels, linestyles="--")


    # Transform the axes appropriately.

    transform = ticker.FuncFormatter(Transform(xmin, xmax, \
            visibilities["image_pixelsize"][index], '%.1f"'))

    ax.set_xticks((ticks[1:-1]-ticks[0])/visibilities["image_pixelsize"][index])
    ax.set_yticks((ticks[1:-1]-ticks[0])/visibilities["image_pixelsize"][index])
    ax.get_xaxis().set_major_formatter(transform)
    ax.get_yaxis().set_major_formatter(transform)

    # Adjust the plot and save it.

    ax.set_xlabel("$\Delta$R.A.", fontsize=fontsize)
    ax.set_ylabel("$\Delta$Dec.", fontsize=fontsize)

    # Adjust the label size as well.

    ax.tick_params(axis="both", which="major", labelsize=fontsize)

    # Return the figure and axes.

    if show_colorbar:
        return fig, ax, cax
    else:
        return fig, ax

# Define a useful class for plotting.

class Transform:
    def __init__(self, xmin, xmax, dx, fmt):
        self.xmin = xmin
        self.xmax = xmax
        self.dx = dx
        self.fmt = fmt

    def __call__(self, x, p):
        return self.fmt% ((x-(self.xmax-self.xmin)/2)*self.dx)



import numpy

def propose_point_emcee(parameters, model="disk"):
    m_env = numpy.random.uniform(-6., parameters["logM_env"]["limits"][1],1)[0]

    # Set up R_env, R_disk, and R_in as they depend on eachother.

    if model == "disk":
        r_env = numpy.random.uniform(max(parameters["logR_env"]["limits"][0], \
                0.5*m_env+4.), parameters["logR_env"]["limits"][1],1)[0]
    elif model == "flared":
        r_env = numpy.random.uniform(parameters["logR_env"]["limits"][0],\
                parameters["logR_env"]["limits"][1],1)[0]

    r_disk = numpy.random.uniform(max(numpy.log10(5.), \
            parameters["logR_disk"]["limits"][0]), min(numpy.log10(500.), \
            r_env, parameters["logR_disk"]["limits"][1]),1)[0]
    r_in = numpy.random.uniform(parameters["logR_in"]["limits"][0],\
            numpy.log10((10.**r_disk)/2),1)[0]

    # Also set up R_cav as it depends on those as well.

    r_cav = numpy.random.uniform(r_in, numpy.log10(0.75*10.**r_disk),1)[0]

    # Same thing for R_gap and w_gap.

    r_gap1 = numpy.random.uniform(numpy.log10(10.**r_in+\
            parameters["w_gap1"]["limits"][0]/2), \
            numpy.log10(0.75*10.**r_disk),1)[0]

    w_gap1 = numpy.random.uniform(parameters["w_gap1"]["limits"][0],\
            min(parameters["w_gap1"]["limits"][1],\
            2*(10.**r_gap1-10.**r_in)), 1)[0]

    # If we are using logTatm0 and logTmid0, they depend on eachother.

    if "logTatm0" in parameters:
        tatm0 = numpy.random.uniform(parameters["logTatm0"]\
                ["limits"][0],parameters["logTatm0"]["limits"][1],1)[0]
        tmid0 = numpy.random.uniform(parameters["logTmid0"]["limits"][0],\
                min(parameters["logTatm0"]["limits"][1], tatm0),1)[0]

    # Loop through and generate the point proposal.

    p = []

    for key in sorted(parameters.keys()):
        if parameters[key]["fixed"]:
            pass
        elif key == "logR_in":
            p.append(r_in)
        elif key == "logR_disk":
            p.append(r_disk)
        elif key == "logR_env":
            p.append(r_env)
        elif key == "logR_cav":
            p.append(r_cav)
        elif key == "logR_gap1":
            p.append(r_gap1)
        elif key == "w_gap1":
            p.append(w_gap1)
        elif key == "logM_disk" and model == "disk":
            p.append(numpy.random.uniform(-6.,parameters[key]["limits"][1], \
                    1)[0])
        elif key == "logM_env":
            p.append(m_env)
        elif key == "h_0":
            p.append(numpy.random.uniform(parameters[key]["limits"][0], \
                    0.2, 1)[0])
        elif key == "logTatm0":
            p.append(tatm0)
        elif key == "logTmid0":
            p.append(tmid0)
        elif key[0:8] == "flux_unc":
            p.append(numpy.random.normal(parameters[key]["value"], 0.001, 1)[0])
        else:
            p.append(numpy.random.uniform(parameters[key]["limits"][0], \
                    parameters[key]["limits"][1], 1)[0])

    # Return the proposed point.

    return p



import warnings

text = warnings.warn("pdspy v2.0.0 represents a major update to the pdspy code, and is not backwards compatible with the results of versions < 2.0.0. *Do not use v2.0.0 to work with results from earlier versions.* For more information, see pdspy.readthedocs.io.", stacklevel=2)

from . import constants
from . import dust
from . import gas
from . import imaging
from . import interferometry
from . import mcmc
from . import misc
from . import modeling
from . import plotting
from . import radmc3d
from . import spectroscopy
from . import stars
from . import statistics
from . import table
from . import utils



import numpy
import scipy.interpolate
from ..constants.astronomy import arcsec
from .libinterferometry import Visibilities
from galario import double
try:
    import trift
except:
    pass

def interpolate_model(u, v, freq, model, nthreads=1, dRA=0., dDec=0., \
        code="galario", nxy=1024, dxy=0.01):

    if code == "galario":
        real = []
        imag = []

        double.threads(nthreads)

        dxy = (model.x[1] - model.x[0])*arcsec

        for i in range(len(model.freq)):
            vis = double.sampleImage(model.image[::-1,:,i,0].copy(order='C'), \
                    dxy, u, v, dRA=dRA*arcsec, dDec=dDec*arcsec)

            real.append(vis.real.reshape((u.size,1)))
            imag.append(-vis.imag.reshape((u.size,1)))

        real = numpy.concatenate(real, axis=1)
        imag = numpy.concatenate(imag, axis=1)

    elif code == "galario-unstructured":
        real = []
        imag = []

        double.threads(nthreads)

        vol = None
        for i in range(len(model.freq)):
            vis = double.sampleUnstructuredImageCPP(model.x*arcsec, \
                    -model.y*arcsec, model.image[:,i].copy(order='C'), nxy, \
                    dxy*arcsec, u, v, dRA=dRA*arcsec, dDec=dDec*arcsec)

            real.append(vis.real.reshape((u.size,1)))
            imag.append(-vis.imag.reshape((u.size,1)))

        real = numpy.concatenate(real, axis=1)
        imag = numpy.concatenate(imag, axis=1)

    elif code == "trift":
        vis = trift.cpu.trift(model.x*arcsec, model.y*arcsec, \
                model.image, u, v, dRA*arcsec, dDec*arcsec, \
                nthreads=nthreads, mode="extended")

        real, imag = vis.real, vis.imag

    return Visibilities(u, v, freq, real, imag, numpy.ones(real.shape))



import numpy
import astropy
import astropy.coordinates

#def match_source_lists(*lists, tol=0.3, table_names=['1', '2']):
def match_source_lists(tol=0.3, table_names=['1', '2'], *lists):

    for i in range(len(lists)-1):
        if i == 0:
            list1 = lists[i]
        else:
            list1 = joined_list
        list2 = lists[i+1]

        coord1 = astropy.coordinates.SkyCoord(list1["ra"].tolist(), \
                list1['dec'].tolist(), frame='icrs')
        coord2 = astropy.coordinates.SkyCoord(list2["ra"].tolist(), \
                list2["dec"].tolist(), frame='icrs')

        idx, d2d, d3d = astropy.coordinates.match_coordinates_sky(coord1,coord2)

        ids, counts = numpy.unique(idx[d2d.arcsec < tol], return_counts=True)
        for j in range(len(ids)):
            if counts[j] > 1:
                min_d2d = min(d2d[idx == ids[j]].arcsec)
                d2d[(d2d.arcsec != min_d2d) & (idx == ids[j])] = \
                        astropy.coordinates.Angle("0d00m{0:f}s".format(2*tol))

        list1["id"] = numpy.arange(len(list1)) + len(list2)
        list2["id"] = numpy.arange(len(list2))
        list1["id"][d2d.arcsec < tol] = idx[d2d.arcsec < tol]

        if i == 0:
            names = table_names[0:2]
        else:
            names = ["",table_names[i+1]]
        
        joined_list = astropy.table.join(list1, list2, join_type='outer', \
                keys='id', table_names=names, \
                uniq_col_name='{col_name}_{table_name}')

        del list1["id"]
        del list2["id"]
        del joined_list["id"]

        if i == 0:
            joined_list["ra"] = numpy.where(\
                    joined_list["ra_"+names[0]].mask, \
                    joined_list["ra_"+names[1]], joined_list["ra_"+names[0]])
            joined_list["dec"] = numpy.where(\
                    joined_list["dec_"+names[0]].mask, \
                    joined_list["dec_"+names[1]], joined_list["dec_"+names[0]])
        else:
            for name in list2.colnames:
                if name in joined_list.colnames:
                    joined_list.rename_column(name, name+"_"+names[1])

            joined_list["ra"] = numpy.where(\
                    joined_list["ra_"].mask, joined_list["ra_"+names[1]], \
                    joined_list["ra_"])
            joined_list["dec"] = numpy.where(\
                    joined_list["dec_"].mask, joined_list["dec_"+names[1]], \
                    joined_list["dec_"])

            del joined_list["ra_"]
            del joined_list["dec_"]

    del joined_list["ra"]
    del joined_list["dec"]

    return joined_list



import numpy

def Ftest(data1, data2, tailed=2):

    N = len(data1)
    M = len(data2)

    mean1 = numpy.mean(data1)
    mean2 = numpy.mean(data2)

    S1 = numpy.sum((data1 - mean1)**2) / (N - 1)
    S2 = numpy.sum((data2 - mean2)**2) / (M - 1)

    F = S1 / S2
    f = numpy.random.f(N-1, M-1, 100000)

    if tailed == 1:
        if F > 1:
            p = len(f[f > F]) / len(f)
        elif F < 1:
            p = len(f[f < F]) / len(f)
    elif tailed == 2:
        if F > 1:
            p = len(f[(f > F) | (f < 1./F)]) / len(f)
        elif F < 1:
            p = len(f[(f > 1./F) | (f < F)]) / len(f)

    return F, p



from .libimaging import Image
from .libimaging import UnstructuredImage
from .imtovis import imtovis
from .readimfits import readimfits
from .readpvfits import readpvfits
from .find import find
from .match_source_lists import match_source_lists
from .update_catalog import update_catalog
from .extract_pv_diagram import extract_pv_diagram



import numpy
from scipy.optimize import brenth
from scipy.integrate import trapz
from ..constants.astronomy import AU, M_sun
from ..constants.math import pi
from ..constants.physics import G, m_p
from ..constants.time import year
from ..dust import Dust
from ..gas import Gas

class UlrichEnvelope:

    def __init__(self, mass=1.0e-3, rmin=0.1, rmax=1000, rcent=30, cavpl=1.0, \
            cavrfact=0.2, t0=None, tpl=None, dust=None, aturb=None):
        self.mass = mass
        self.rmin = rmin
        self.rmax= rmax
        self.rcent = rcent
        self.cavpl = cavpl
        self.cavrfact = cavrfact
        self.t0 = t0
        self.tpl = tpl
        if (dust != None):
            self.dust = dust

        self.gas = []
        self.abundance = []
        self.aturb = aturb

    def add_gas(self, gas, abundance):
        self.gas.append(gas)
        self.abundance.append(abundance)

    def density(self, r, theta, phi, code="hyperion"):
        #numpy.seterr(all='ignore')
        ##### Star parameters

        mstar = M_sun
        
        ##### Envelope parameters

        rin = self.rmin * AU
        rout = self.rmax * AU
        mass = self.mass * M_sun
        rcent = self.rcent * AU
        cavz0 = 1*AU
        cavpl = self.cavpl
        cavrfact = self.cavrfact

        # Set up the coordinates.
        
        rr, tt, pp = numpy.meshgrid(r*AU, theta, phi,indexing='ij')

        mu = numpy.cos(tt)

        RR = rr * numpy.sin(tt)
        zz = rr * numpy.cos(tt)

        # Calculate mu0 at each r, theta combination.

        def func(mu0,r,mu,R_c):
            return mu0**3-mu0*(1-r/R_c)-mu*(r/R_c)

        mu0 = mu*0.
        for ir in range(rr.shape[0]):
            for it in range(rr.shape[1]):
                mu0[ir,it,0] = brenth(func,-1.0,1.0,args=(rr[ir,it,0], \
                        mu[ir,it,0],rcent))

        ##### Make the dust density model for an Ulrich envelope.

        rho0 = 1.0
        
        rho = rho0 * (rr / rcent)**(-1.5) * (1 + mu/mu0)**(-0.5)* \
                (mu/mu0 + 2*mu0**2 * rcent/rr)**(-1)

        mid1 = (numpy.abs(mu) < 1.0e-10) & (rr < rcent)
        rho[mid1] = rho0 * (rr[mid1] / rcent)**(-0.5) * \
                (1. - rr[mid1] / rcent)**(-1) / 2.

        mid2 = (numpy.abs(mu) < 1.0e-10) & (rr > rcent)
        rho[mid2] = rho0 * (2.*rr[mid2]/rcent - 1)**(-0.5) * \
                (rr[mid2]/rcent - 1.)**(-1)

        rho[(rr >= rout) ^ (rr <= rin)] = 0e0

        ##### Add an outflow cavity.

        rho[numpy.abs(zz)/AU-cavz0/AU-(RR/AU)**cavpl > 0.0] *= cavrfact
        
        ##### Normalize the mass correctly.
        
        if theta.max() > pi/2:
            mdot = mass/(2*pi*trapz(trapz(rho*rr**2*numpy.sin(tt),tt,axis=1), \
                    rr[:,0,:],axis=0))[0]
        else:
            mdot = mass/(4*pi*trapz(trapz(rho*rr**2*numpy.sin(tt),tt,axis=1), \
                    rr[:,0,:],axis=0))[0]
        rho *= mdot

        #numpy.seterr(all='warn')

        return rho

    def temperature(self, r, theta, phi):
        ##### Disk Parameters
        
        rin = self.rmin * AU
        rout = self.rmax * AU
        t0 = self.t0
        tpl = self.tpl

        ##### Set up the coordinates

        rt, tt, pp = numpy.meshgrid(r*AU, theta, phi,indexing='ij')

        rr = rt*numpy.sin(tt)
        zz = rt*numpy.cos(tt)

        ##### Make the dust density model for a protoplanetary disk.
        
        t = t0 * (rt / (1*AU))**(-tpl)

        t[(rr >= rout) ^ (rr <= rin)] = 0.1

        t[t > 10000.] = 10000.
        
        return t

    def number_density(self, r, theta, phi, gas=0):
        rho = self.density(r, theta, phi)

        rho_gas = rho * 100

        n_H2 = rho_gas * 0.8 / (2.37*m_p)

        n = n_H2 * self.abundance[gas]

        return n

    def microturbulence(self, r, theta, phi):
        ##### Set up the coordinates

        rt, tt, pp = numpy.meshgrid(r*AU, theta, phi,indexing='ij')

        rr = rt*numpy.sin(tt)
        zz = rt*numpy.cos(tt)

        ##### Make the dust density model for a protoplanetary disk.
        
        aturb = numpy.ones(rr.shape)*self.aturb*1.0e5
        
        return aturb

    def velocity(self, r, theta, phi, mstar=0.5):
        mstar *= M_sun
        rcent = self.rcent * AU

        # Set up the coordinates.
        
        rr, tt, pp = numpy.meshgrid(r*AU, theta, phi,indexing='ij')

        mu = numpy.cos(tt)

        # Calculate mu0 at each r, theta combination.

        def func(mu0,r,mu,R_c):
            return mu0**3-mu0*(1-r/R_c)-mu*(r/R_c)

        mu0 = mu*0.
        for ir in range(rr.shape[0]):
            for it in range(rr.shape[1]):
                mu0[ir,it,0] = brenth(func,-1.0,1.0,args=(rr[ir,it,0], \
                        mu[ir,it,0],rcent))

        v_r = -numpy.sqrt(G*mstar/rr)*numpy.sqrt(1 + mu/mu0)
        v_theta = numpy.sqrt(G*mstar/rr) * (mu0 - mu) * \
                numpy.sqrt((mu0 + mu) / (mu0 * numpy.sin(tt)**2))
        v_phi = numpy.sqrt(G*mstar/rr) * numpy.sqrt((1 - mu0**2)/(1 - mu**2)) *\
                numpy.sqrt(1 - mu/mu0)

        return numpy.array((v_r, v_theta, v_phi))

    def calculate_mdot(self, r, theta, phi, mstar=0.5, code="hyperion"):
        #numpy.seterr(all='ignore')
        ##### Star parameters

        mstar *= M_sun
        
        ##### Envelope parameters

        rin = self.rmin * AU
        rout = self.rmax * AU
        mass = 100 * self.mass * M_sun
        rcent = self.rcent * AU
        cavz0 = 1*AU
        cavpl = self.cavpl
        cavrfact = self.cavrfact

        # Set up the coordinates.
        
        rr, tt, pp = numpy.meshgrid(r*AU, theta, phi,indexing='ij')

        mu = numpy.cos(tt)

        RR = rr * numpy.sin(tt)
        zz = rr * numpy.cos(tt)

        # Calculate mu0 at each r, theta combination.

        def func(mu0,r,mu,R_c):
            return mu0**3-mu0*(1-r/R_c)-mu*(r/R_c)

        mu0 = mu*0.
        for ir in range(rr.shape[0]):
            for it in range(rr.shape[1]):
                mu0[ir,it,0] = brenth(func,-1.0,1.0,args=(rr[ir,it,0], \
                        mu[ir,it,0],rcent))

        ##### Make the dust density model for an Ulrich envelope.

        rho0 = 1.0
        
        rho = rho0 * (rr / rcent)**(-1.5) * (1 + mu/mu0)**(-0.5)* \
                (mu/mu0 + 2*mu0**2 * rcent/rr)**(-1)

        mid1 = (numpy.abs(mu) < 1.0e-10) & (rr < rcent)
        rho[mid1] = rho0 * (rr[mid1] / rcent)**(-0.5) * \
                (1. - rr[mid1] / rcent)**(-1) / 2.

        mid2 = (numpy.abs(mu) < 1.0e-10) & (rr > rcent)
        rho[mid2] = rho0 * (2.*rr[mid2]/rcent - 1)**(-0.5) * \
                (rr[mid2]/rcent - 1.)**(-1)

        rho[(rr >= rout) ^ (rr <= rin)] = 0e0

        ##### Normalize the mass correctly.
        
        if theta.max() > pi/2:
            prefix = mass/(2*pi*trapz(trapz(rho*rr**2*numpy.sin(tt),tt,axis=1),\
                    rr[:,0,:],axis=0))[0]
        else:
            prefix = mass/(4*pi*trapz(trapz(rho*rr**2*numpy.sin(tt),tt,axis=1),\
                    rr[:,0,:],axis=0))[0]

        mdot = prefix * (4*numpy.pi * numpy.sqrt(G * mstar) * rcent**1.5) / \
                (M_sun / year)

        return mdot


    def read(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "r")
        else:
            f = usefile

        self.mass = f['mass'][()]
        self.rmin = f['rmin'][()]
        self.rmax = f['rmax'][()]
        self.rcent = f['rcent'][()]
        self.cavpl = f['cavpl'][()]
        self.cavrfact = f['cavrfact'][()]

        if 't0' in f:
            self.t0 = f['t0'][()]
            self.tpl = f['tpl'][()]

        if 'aturb' in f:
            self.aturb = f['aturb'][()]

        if ('Dust' in f):
            self.dust = Dust()
            self.dust.set_properties_from_file(usefile=f['Dust'])

        for name in f['Gas']:
            self.gas.append(Gas())
            self.abundance.append(f['Gas'][name]['Abundance'][()])
            self.gas[-1].set_properties_from_file(usefile=f['Gas'][name])

        if (usefile == None):
            f.close()

    def write(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "w")
        else:
            f = usefile

        f['mass'] = self.mass
        f['rmin'] = self.rmin
        f['rmax'] = self.rmax
        f['rcent'] = self.rcent
        f['cavpl'] = self.cavpl
        f['cavrfact'] = self.cavrfact

        if self.t0 != None:
            f['t0'] = self.t0
            f['tpl'] = self.plt

        if self.aturb != None:
            f['aturb'] = self.aturb

        if hasattr(self, 'dust'):
            dust = f.create_group("Dust")
            self.dust.write(usefile=dust)

        gases = []
        if hasattr(self, 'gas'):
            gas = f.create_group("Gas")
            for i in range(len(self.gas)):
                gases.append(gas.create_group("Gas{0:d}".format(i)))
                gases[i]["Abundance"] = self.abundance[i]
                self.gas[i].write(usefile=gases[i])

        if (usefile == None):
            f.close()



from .. import interferometry as uv
from .. import spectroscopy as sp
from .. import imaging as im
import numpy

def load_data(config, model="disk", gridsize1D=20):
    r"""
    Loads the data, as specified by the configuration module.

    Args:
        :attr:`config` (`module`):
            The configuration information for the model, as read in by the :code:utils.load_config function.
        :attr:`model` (`str`, optional):
            Are you running disk_model_* or flared_model_*. Default: `"disk"`
        :attr:`model` (`int` or `list-like`, optional):
            The number of bins to use when averaging the visibility data into 1D. Default: `40`

    Returns:
        :attr:`visibilities` (`dict`):
            A dictionary containing the visibility data and the information about it from the configuration module.
        :attr:`images` (`dict`):
            A dictionary containing the image data and the information about it from the configuration module.
        :attr:`spectra` (`dict`):
            A dictionary containing the spectra and the information about it from the configuration module.
    """


    # Set up the places where we will put all of the data.

    config.visibilities["data"] = []
    config.visibilities["data1d"] = []
    config.visibilities["data2d"] = []
    config.visibilities["image"] = []
    config.spectra["data"] = []
    config.spectra["binned"] = []
    config.images["data"] = []

    # Make sure gridsize1D has the proper number of elements.

    if gridsize1D == 20:
        gridsize1D = [20 for i in range(len(config.visibilities["file"]))]

    # Read in the millimeter visibilities.

    for j in range(len(config.visibilities["file"])):
        # Read the raw data.

        if config.visibilities["file"][j].split(".")[-1] == "ms":
            data = uv.readms(config.visibilities["file"][j], \
                    spw=config.visibilities["spw"][j], \
                    tolerance=config.visibilities["tolerance"][j], \
                    datacolumn=config.visibilities["datacolumn"][j])
        else:
            data = uv.Visibilities()
            data.read(config.visibilities["file"][j])

        # Center the data. => need to update!

        data = uv.center(data, [config.visibilities["x0"][j], \
                config.visibilities["y0"][j], 1.])

        # Set any weights < 0 to be equal to zero to prevent odd fits.

        data.weights[data.weights < 0] = 0

        # Take the complex conjugate to make sure orientation is correct.

        #data.imag *= -1

        # Add the data to the dictionary structure.

        config.visibilities["data"].append(data)

        # Scale the weights of the visibilities to force them to be fit well.

        config.visibilities["data"][j].weights *= \
                config.visibilities["weight"][j]

        # Average the visibilities radially.

        if model == "disk":
            config.visibilities["data1d"].append(uv.average(data, \
                    gridsize=gridsize1D[j], radial=True, log=True, \
                    logmin=data.uvdist[numpy.nonzero(data.uvdist)].min()*0.95, \
                    logmax=data.uvdist.max()*1.05))
        elif model == "flared":
            config.visibilities["data1d"].append(uv.average(data, \
                    gridsize=gridsize1D[j], radial=True, log=True, \
                    logmin=data.uvdist[numpy.nonzero(data.uvdist)].min()*0.95, \
                    logmax=data.uvdist.max()*1.05, mode="spectralline"))

        # Also average the data into a 2D grid, if disk model..

        if model == "disk":
            config.visibilities["data2d"].append(uv.grid(data, \
                    gridsize=config.visibilities["gridsize"][j], \
                    binsize=config.visibilities["binsize"][j]))

        # Read in the image.

        config.visibilities["image"].append(im.readimfits(\
                config.visibilities["image_file"][j]))

    ######################
    # Read in the spectra.
    ######################

    for j in range(len(config.spectra["file"])):
        config.spectra["data"].append(sp.Spectrum())
        config.spectra["data"][j].read(config.spectra["file"][j])

        # Adjust the weight of the SED, as necessary.

        config.spectra["data"][j].unc /= config.spectra["weight"][j]**0.5

        # Merge the SED with the binned Spitzer spectrum.

        if config.spectra["bin?"][j]:
            wave = numpy.linspace(config.spectra["data"][j].wave.min(), \
                    config.spectra["data"][j].wave.max(), \
                    config.spectra["nbins"][j])
            flux = numpy.interp(wave, config.spectra["data"][j].wave, \
                    config.spectra["data"][j].flux)

            good = flux > 0.

            wave = wave[good]
            flux = flux[good]

            config.spectra["binned"].append(sp.Spectrum(wave, flux))
        else:
            config.spectra["binned"].append(config.spectra["data"][j])

        # Merge all the spectra together in one big SED.

        const_unc = 0.1 / config.spectra["weight"][j]**0.5

        try:
            config.spectra["total"].wave = numpy.concatenate((\
                    config.spectra["total"].wave, \
                    config.spectra["binned"][j].wave))
            config.spectra["total"].flux = numpy.concatenate((\
                    config.spectra["total"].flux, \
                    numpy.log10(config.spectra["binned"][j].flux)))
            config.spectra["total"].unc = numpy.concatenate((\
                    config.spectra["total"].unc, \
                    numpy.repeat(const_unc, \
                    config.spectra["binned"][j].wave.size)))

            order = numpy.argsort(config.spectra["total"].wave)

            config.spectra["total"].wave = config.spectra["total"].wave[order]
            config.spectra["total"].flux = config.spectra["total"].flux[order]
            config.spectra["total"].unc = config.spectra["total"].unc[order]
        except:
            config.spectra["total"] = sp.Spectrum(\
                    config.spectra["binned"][j].wave, \
                    numpy.log10(config.spectra["binned"][j].flux), \
                    numpy.repeat(const_unc, \
                    config.spectra["binned"][j].wave.size))

    #####################
    # Read in the images.
    #####################

    for j in range(len(config.images["file"])):
        config.images["data"].append(im.Image())
        config.images["data"][j].read(config.images["file"][j])

    return config.visibilities, config.images, config.spectra



from .Disk import Disk
from .DartoisDisk import DartoisDisk
from .SettledDisk import SettledDisk

from .PringleDisk import PringleDisk
from .DartoisPringleDisk import DartoisPringleDisk
from .SettledPringleDisk import SettledPringleDisk

from .Grid import Grid
from .Model import Model
from .Star import Star
from .UlrichEnvelope import UlrichEnvelope
from .UlrichEnvelopeExtended import UlrichEnvelopeExtended
from .TaperedUlrichEnvelope import TaperedUlrichEnvelope
from .TaperedUlrichEnvelopeExtended import TaperedUlrichEnvelopeExtended
from .YSOModel import YSOModel

from .run_disk_model import run_disk_model
from .run_flared_model import run_flared_model
from .check_parameters import check_parameters
from .get_surrogate_model import get_surrogate_model



from .libinterferometry import Visibilities, average, grid, freqcorrect, chisq
from .readuvfits import readuvfits
from .readvis import readvis
from .center import center
from .clean import clean
from .concatenate import concatenate
from .fit_model import fit_model
from .invert import invert
from .model import model
from .rotate import rotate
from .interpolate_model import interpolate_model

try:
    from .readms import readms
except:
    pass

from .rmlimage import rmlimage



import scipy.interpolate
import astropy.table
import numpy
import time
import glob

############################################################################
#
# Read the data files into a table.
#
############################################################################

def read_pms_data(tracks="BHAC15"):

    # Load in the data for the appropriate set of evolutionary tracks.

    path = '/'.join(__file__.split("/")[0:-1])+"/evolutionary_tracks/"
    
    if tracks == "BHAC15":
        f = open(path+"BHAC15_tracks+structure.txt","r")
        lines = f.readlines()
        f.close()

        colnames = lines[46].split()[1:]

        data = numpy.loadtxt(path+"BHAC15_tracks+structure.txt", comments="!", \
                skiprows=45)
    elif tracks == "Siess2000":
        f = open(path+"siess_2000/m0.13z02.hrd")
        lines = f.readlines()
        f.close()

        line1 = lines[0].replace(" (","_(").replace("log g","logg").\
                replace("#","").replace("_(Lo)","/Ls").replace("age_","log_t").\
                split()
        line2 = lines[1].replace(" (","_(").replace("log g","logg").\
                replace("#","").replace("_(Mo)","/Ms").split()

        colnames = line1 + line2
        colnames[0::2] = line1
        colnames[1::2] = line2

        files = glob.glob(path+"siess_2000/*.hrd")
        for file in files:
            try:
                data = numpy.concatenate((data, numpy.loadtxt(file)))
            except:
                data = numpy.loadtxt(file)

        # Fix the stellar luminosity.

        data[:,2] = numpy.log10(data[:,2])
        data[:,-1] = numpy.log10(data[:,-1])
    elif tracks == "Dotter2008":
        f = open(path+"dotter2008/m200fehp00afep0.trk")
        lines = f.readlines()
        f.close()

        # Get the column names.

        colnames = ['M/Ms'] + lines[1].replace("Log ","Log").\
                replace("Age ","log_t").replace("LogT","Teff").\
                replace("LogL","L/Ls").replace("yrs","yr").split()[1:]

        # Now read in the data files.

        files = glob.glob(path+"dotter2008/*.trk")
        for file in files:
            new_data = numpy.loadtxt(file)

            # add a column with the mass of the star.

            mass = float(file.split("/")[-1][1:4])/100.
            mass_arr = numpy.zeros((new_data.shape[0],1)) + mass
            new_data = numpy.hstack((mass_arr, new_data))

            # Merge with existing data.

            try:
                data = numpy.concatenate((data, new_data))
            except:
                data = new_data.copy()

        # Get rid of ages more than ~50Myr.

        good = data[:,1] < 50.0e6
        data = data[good,:]

        # Fix some of the columns.

        data[:,1] = numpy.log10(data[:,1])
        data[:,2] = 10.**data[:,2]

    elif tracks == "Tognelli2011":
        f = open(path+"tognelli2011/Z0.02000_Y0.2700_XD2E5_ML1.68_AS05/"
                "TRK_M0.20_Z0.02000_Y0.2700_XD2E5_ML1.68_AS05.DAT")
        lines = f.readlines()
        f.close()

        # Get the column names.

        colnames = ['M/Ms'] + lines[3].replace("LOG ","LOG_").\
                replace("LOG_AGE","log_t(yr)").replace("LOG_L","L/Ls").\
                replace("LOG_TE","Teff").split()[1:]

        # Now read in the data files.

        files = glob.glob(path+"tognelli2011/"
                "Z0.02000_Y0.2700_XD2E5_ML1.68_AS05/*.DAT")
        for file in files:
            new_data = numpy.loadtxt(file)

            # add a column with the mass of the star.

            mass = float(file.split("/")[-1].split("_")[1][1:])
            mass_arr = numpy.zeros((new_data.shape[0],1)) + mass
            new_data = numpy.hstack((mass_arr, new_data))

            # Merge with existing data.

            try:
                data = numpy.concatenate((data, new_data))
            except:
                data = new_data.copy()

        # Fix some of the columns.

        data[:,5] = 10.**data[:,5]

    elif tracks == "Feiden2016":
        f = open(path+"feiden2016/std/"
                "m0090_GS98_p000_p0_y28_mlt1.884.trk")
        lines = f.readlines()
        f.close()

        # Get the column names.

        colnames = ['M/Ms'] + lines[3].replace("Log ","Log").\
                replace("Age ","log_t").replace("LogT","Teff").\
                replace("LogL","L/Ls").replace("yrs","yr").split()[1:6]

        # Now read in the data files.

        files = glob.glob(path+"feiden2016/std/*.trk")
        for file in files:
            new_data = numpy.loadtxt(file, usecols=(0,1,2,3,4))

            # add a column with the mass of the star.

            mass = float(file.split("/")[-1][1:5])/1000.
            mass_arr = numpy.zeros((new_data.shape[0],1)) + mass
            new_data = numpy.hstack((mass_arr, new_data))

            # Merge with existing data.

            try:
                data = numpy.concatenate((data, new_data))
            except:
                data = new_data.copy()

        # Get rid of ages more than ~50Myr.

        good = data[:,1] < 50.0e6
        data = data[good,:]

        # Fix some of the columns.

        data[:,1] = numpy.log10(data[:,1])
        data[:,2] = 10.**data[:,2]

    elif tracks == "Feiden2016mag":
        f = open(path+"feiden2016/mag/"
                "m1700_GS98_p000_p0_y28_mlt1.884_mag08kG.ntrk")
        lines = f.readlines()
        f.close()

        # Get the column names.

        colnames = ['M/Ms'] + lines[8].replace("conv. ","conv.").\
                replace("AGE","log_t(yr)").replace("log(Teff)","Teff").\
                replace("log(L/Lsun)","L/Ls").replace("Model #","Model#").\
                replace("M He core","M_He_core").replace(",","").split()[1:]

        # Now read in the data files.

        files = glob.glob(path+"feiden2016/mag/*.ntrk")
        for file in files:
            new_data = numpy.loadtxt(file,usecols=tuple([i for i in range(12)]))

            # add a column with the mass of the star.

            mass = float(file.split("/")[-1][1:5])/1000.
            mass_arr = numpy.zeros((new_data.shape[0],1)) + mass
            new_data = numpy.hstack((mass_arr, new_data))

            # Merge with existing data.

            try:
                data = numpy.concatenate((data, new_data))
            except:
                data = new_data.copy()

        # Fix some of the columns.

        data[:,3] = numpy.log10(data[:,3]*1.0e9)
        data[:,7] = 10.**data[:,7]

    elif tracks == "Chen2014":
        f = open(path+"bressan2012/Z0.017Y0.279/"
                "Z0.017Y0.279OUTA1.77_F7_M000.700.DAT")
        lines = f.readlines()
        f.close()

        # Get the column names.

        colnames = lines[0].replace("LOG ","LOG_").\
                replace("AGE","log_t(yr)").replace("LOG_L","L/Ls").\
                replace("LOG_TE","Teff").replace("MASS","M/Ms").split()

        # Now read in the data files.

        files = glob.glob(path+"bressan2012/Z0.017Y0.279/*.DAT")

        for file in files:
            new_data = numpy.loadtxt(file, skiprows=2)

            # Merge with existing data.

            try:
                data = numpy.concatenate((data, new_data))
            except:
                data = new_data.copy()

        # Get rid of ages more than ~50Myr.

        good = data[:,2] < 50.0e6
        data = data[good,:]

        # Fix some of the columns.

        data[:,2] = numpy.log10(data[:,2])
        data[:,4] = 10.**data[:,4]

    elif tracks == "Bressan2012":
        f = open(path+"bressan2012/bressan2012.dat")
        lines = f.readlines()
        f.close()

        # Get the column names.

        colnames = lines[13].replace("log(age/yr)","log_t(yr)").\
                replace("logL/Lo","L/Ls").replace("logTe","Teff").\
                replace("M_act","M/Ms").split()[1:]

        # Now read in the data files.

        data = numpy.loadtxt(path+"bressan2012/bressan2012.dat")

        # Fix some of the columns.

        data[:,5] = 10.**data[:,5]

    # Make the data into a table.

    table = astropy.table.Table(data, names=colnames)

    # Return the table now.

    return table

############################################################################
#
# Get Mass from Teff and Lstar
#
############################################################################

def pms_get_mstar(temperature=None, luminosity=None, tracks="BHAC15"):

    # Load in the data for the appropriate set of evolutionary tracks.

    table = read_pms_data(tracks=tracks)

    # Now do the 2D interpolation.

    Mstar = scipy.interpolate.LinearNDInterpolator((table["Teff"], \
            table["L/Ls"]), table["M/Ms"], rescale=True)

    # Finally, get the stellar mass.

    if isinstance(temperature,float) and isinstance(luminosity,float):
        xi = numpy.array([[temperature, numpy.log10(luminosity)]])
    else:
        xi = numpy.array([[temperature[i],numpy.log10(luminosity[i])] for i in \
                range(len(temperature))])

    return Mstar(xi)

############################################################################
#
# Get Age from Teff and Lstar
#
############################################################################

def pms_get_age(temperature=None, luminosity=None, tracks="BHAC15"):

    # Load in the data for the appropriate set of evolutionary tracks.

    table = read_pms_data(tracks=tracks)

    # Now do the 2D interpolation.

    Age = scipy.interpolate.LinearNDInterpolator((table["Teff"], \
            table["L/Ls"]), table["log_t(yr)"], rescale=True)

    # Finally, get the stellar mass.

    if isinstance(temperature,float) and isinstance(luminosity,float):
        xi = numpy.array([[temperature, numpy.log10(luminosity)]])
    else:
        xi = numpy.array([[temperature[i],numpy.log10(luminosity[i])] for i in \
                range(len(temperature))])

    return 10.**Age(xi)

############################################################################
#
# Get Teff from Mass and Age.
#
############################################################################

def pms_get_teff(mass=1.0, age=1.0e6, tracks="BHAC15"):

    # Load in the data for the appropriate set of evolutionary tracks.

    table = read_pms_data(tracks=tracks)

    # Now do the 2D interpolation.

    Teff = scipy.interpolate.LinearNDInterpolator((table["M/Ms"], \
            table["log_t(yr)"]), table["Teff"])

    # Finally, get the stellar mass.

    if isinstance(age,float) and isinstance(mass,float):
        xi = numpy.array([[mass, numpy.log10(age)]])
    elif isinstance(age,float):
        xi = numpy.array([[mass[i],numpy.log10(age)] for i in range(len(mass))])
    else:
        xi = numpy.array([[mass,numpy.log10(age[i])] for i in range(len(age))])

    return Teff(xi)

############################################################################
#
# Get Lstar from Mass and Age.
#
############################################################################

def pms_get_luminosity(mass=1.0, age=1.0e6, tracks="BHAC15"):

    # Load in the data for the appropriate set of evolutionary tracks.

    table = read_pms_data(tracks=tracks)

    # Now do the 2D interpolation.

    Lstar = scipy.interpolate.LinearNDInterpolator((table["M/Ms"],\
            table["log_t(yr)"]), table["L/Ls"])

    # Finally, get the stellar mass.

    if isinstance(age,float) and isinstance(mass,float):
        xi = numpy.array([[mass, numpy.log10(age)]])
    elif isinstance(age,float) and not isinstance(mass,float):
        xi = numpy.array([[mass[i],numpy.log10(age)] for i in range(len(mass))])
    elif not isinstance(age,float) and isinstance(mass,float):
        xi = numpy.array([[mass,numpy.log10(age[i])] for i in range(len(age))])
    else:
        xi = numpy.array([[mass[i],numpy.log10(age[i])] for i in range(len(age))])

    return 10.**Lstar(xi)

############################################################################
#
# Get Rstar from Mass and Age.
#
############################################################################

def pms_get_radius(mass=1.0, age=1.0e6, tracks="BHAC15"):

    # Load in the data for the appropriate set of evolutionary tracks.

    table = read_pms_data(tracks=tracks)

    # Now do the 2D interpolation.

    Rstar = scipy.interpolate.LinearNDInterpolator((table["M/Ms"],\
            table["log_t(yr)"]), table["LogR"])

    # Finally, get the stellar mass.

    if isinstance(age,float) and isinstance(mass,float):
        xi = numpy.array([[mass, numpy.log10(age)]])
    elif isinstance(age,float) and not isinstance(mass,float):
        xi = numpy.array([[mass[i],numpy.log10(age)] for i in range(len(mass))])
    elif not isinstance(age,float) and isinstance(mass,float):
        xi = numpy.array([[mass,numpy.log10(age[i])] for i in range(len(age))])
    else:
        xi = numpy.array([[mass[i],numpy.log10(age[i])] for i in range(len(age))])

    return 10.**Rstar(xi)



import numpy
import scipy
import h5py
from .Dust import Dust

class DustGenerator:
    def __init__(self, dust, with_dhs=False, fmax=0.8, nf=50, singlesize=False):
        if type(dust) == str:
            self.read(dust)
        else:
            self.rho = dust.rho

            self.amax = numpy.logspace(-4.,1.,60)
            self.p = numpy.linspace(2.5,4.5,11)
            self.lam = dust.lam

            self.kabs = []
            self.ksca = []
            self.kext = []
            self.albedo = []

            a = numpy.logspace(numpy.log10(0.05e-4),1.,500)
            if singlesize:
                self.amax = a
            kabsgrid = numpy.zeros((self.lam.size, a.size))
            kscagrid = numpy.zeros((self.lam.size, a.size))

            for i in range(a.size):
                if with_dhs:
                    dust.calculate_dhs_opacity(a[i], fmax=fmax, nf=nf, nang=1)
                else:
                    dust.calculate_opacity(a[i], coat_volume_fraction=0.0, \
                            nang=1)

                kabsgrid[:,i] = dust.kabs
                kscagrid[:,i] = dust.ksca

            for p in self.p:
                kabs_temp = []
                ksca_temp = []
                kext_temp = []
                albedo_temp = []

                for amax in self.amax:
                    if singlesize:
                        normfunc = numpy.zeros(a.size)
                        normfunc[a == amax] = 1.
                    else:
                        normfunc = a**(3-p)
                        normfunc[a > amax] = 0.

                    norm = scipy.integrate.trapz(normfunc, x=a)

                    kabs_temp.append(scipy.integrate.trapz(kabsgrid*normfunc,\
                            x=a, axis=1)/norm)
                    ksca_temp.append(scipy.integrate.trapz(kscagrid*normfunc,\
                            x=a, axis=1)/norm)
                    kext_temp.append(kabs_temp[-1] + ksca_temp[-1])
                    albedo_temp.append( ksca_temp[-1] / kext_temp[-1])

                self.kabs.append(kabs_temp)
                self.ksca.append(ksca_temp)
                self.kext.append(kext_temp)
                self.albedo.append(albedo_temp)

            self.kabs = numpy.array(self.kabs)
            self.ksca = numpy.array(self.ksca)
            self.kext = numpy.array(self.kext)
            self.albedo = numpy.array(self.albedo)

    def __call__(self, amax, p=None):
        if self.old:
            f_kabs = scipy.interpolate.interp2d(self.lam, self.amax, \
                    numpy.log10(self.kabs))
            f_ksca = scipy.interpolate.interp2d(self.lam, self.amax, \
                    numpy.log10(self.ksca))

            kabs = 10.**f_kabs(self.lam, amax)
            ksca = 10.**f_ksca(self.lam, amax)

            d = Dust()
            d.set_properties(self.lam, kabs, ksca)
        else:
            f_kabs = scipy.interpolate.RegularGridInterpolator(\
                    (self.p, self.amax, self.lam), numpy.log10(self.kabs))
            f_ksca = scipy.interpolate.RegularGridInterpolator(\
                    (self.p, self.amax, self.lam), numpy.log10(self.ksca))

            pts = numpy.array([[p, amax, lam] for lam in self.lam])

            kabs = 10.**f_kabs(pts)
            ksca = 10.**f_ksca(pts)

            d = Dust()
            d.set_properties(self.lam, kabs, ksca)

        return d

    def read(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "r")
        else:
            f = usefile

        self.lam = f['lam'][...]
        self.amax = f['amax'][...]

        if ('p' in f):
            self.p = f['p'][...]
            self.old = False
        else:
            self.old = True

        if ('kabs' in f):
            self.kabs = f['kabs'][...]
        if ('ksca' in f):
            self.ksca = f['ksca'][...]
        if (hasattr(self, 'kabs') and hasattr(self, 'ksca')):
            self.kext = self.kabs + self.ksca
            self.albedo = self.ksca / self.kext

        if ('rho' in f):
            self.rho = f['rho'][...][0]

        if (usefile == None):
            f.close()

    def write(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "w")
        else:
            f = usefile

        if hasattr(self, 'amax'):
            amax_dset = f.create_dataset("amax", (self.amax.size,), dtype='f')
            amax_dset[...] = self.amax
        if hasattr(self, 'p'):
            amax_dset = f.create_dataset("p", (self.p.size,), dtype='f')
            amax_dset[...] = self.p
        if hasattr(self, 'lam'):
            lam_dset = f.create_dataset("lam", (self.lam.size,), dtype='f')
            lam_dset[...] = self.lam
        
        if hasattr(self, 'kabs'):
            kabs_dset = f.create_dataset("kabs", self.kabs.shape, dtype='f')
            kabs_dset[...] = self.kabs
        if hasattr(self, 'ksca'):
            ksca_dset = f.create_dataset("ksca", self.ksca.shape, dtype='f')
            ksca_dset[...] = self.ksca
        if hasattr(self, 'g'):
            g_dset = f.create_dataset("g", self.g.shape, dtype='f')
            g_dset[...] = self.g

        if hasattr(self, 'rho'):
            rho_dset = f.create_dataset("rho", (1,), dtype='f')
            rho_dset[...] = [self.rho]

        if (usefile == None):
            f.close()



from .Dust import Dust
import os
import numpy

def run_opacity_tool(amin=0.05, amax=3000., apow=3.5, na=50, fcarbon=0.13, \
        Vcarbon=0.15, porosity=0.25, fmax=0.8, lmin=0.05, lmax=5000, nlam=300, \
        filename=None, verbose=False):

    original_dir = os.environ["PWD"]
    os.mkdir("/tmp/temp_opacitytool")
    os.chdir("/tmp/temp_opacitytool")

    try:
        command = "OpacityTool -amin {0:f} -amax {1:f} -apow {2:f} -na {3:d} "\
                "-fcarbon {4:f} -Vcarbon {5:f} -porosity {6:f} -fmax {7:f} "\
                "-lmin {8:f} -lmax {9:f} -nlam {10:d}".format(amin, amax, apow,\
                na, fcarbon, Vcarbon, porosity, fmax, lmin, lmax, nlam)

        if not verbose:
            os.system(command + " > log.txt")
        else:
            os.system(command)
        
        data = numpy.loadtxt("particle.dat")

        d = Dust()
        d.set_properties(data[:,0], data[:,1], data[:,2])

        if not verbose:
            os.system("rm particle.dat log.txt")
        else:
            os.system("rm particle.dat")
        os.chdir(original_dir)
        os.rmdir("/tmp/temp_opacitytool")

        if filename != None:
            d.write(filename)
        else:
            return d
    except:
        os.system("rm particle.dat log.txt")
        os.chdir(original_dir)
        os.rmdir("/tmp/temp_opacitytool")



from ..interferometry import Visibilities, clean, average
from ..constants.physics import c
from matplotlib.colors import LinearSegmentedColormap
import matplotlib.patches as patches
import matplotlib.ticker as ticker
import matplotlib.pyplot as plt
import numpy

def plot_channel_maps(visibilities, model, parameters, params, index=0, \
        plot_vis=False, fig=None, image="data", contours="model", \
        model_image="beam-convolve", maxiter=100, threshold=1., uvtaper=None, \
        weighting="natural", robust=2.0, \
        vmin=None, vmax=None, levels=None, negative_levels=None, \
        image_cmap="viridis", contours_colors=None, fontsize="medium", \
        show_velocity=True, show_beam=True, vis_color="b", vis_model_color="g",\
        show_xlabel=True, show_ylabel=True, skip=0, \
        auto_center_velocity=False, v_width=10., beamxy=(0.15,0.15), \
        show_colorbar=False, cax=None, colorbar_location='right', \
        colorbar_orientation='vertical', colorbar_size='10%', \
        colorbar_pad=0.01, units="Jy/beam", vis_marker="o"):
    r"""
    Plot the millimeter channel maps, along with the best fit model.

    Args:
        :attr:`visibilities` (`dict`):
            Dictionary containing the visibility data, typically as loaded by 
            :code:`utils.load_config` and :code:`utils.load_data`.
        :attr:`model` (`modeling.Model`):
            The radiative transfer model that you would like to plot the 
            visibilities of. Typically this is the output of 
            modeling.run_disk_model.
        :attr:`parameters` (`dict`):
            The parameters dictionary in the config module as loaded in by 
            :code:`utils.load_config`
        :attr:`params` (`dict`):
            The parameters of the model, typically as a dictionary mapping 
            parameter keys from the :code:`parameters` dictionary to their 
            values.
        :attr:`index` (`int`, optional):
            The visibilities dictionary typically contains a list of datasets. 
            `index` indicates which one to plot.
        :attr:`plot_vis` (`bool`, optional):
            If `True`, plot the azimuthally averaged visibilities instead of 
            images. Defautl: `False`
        :attr:`fig` (`tuple`, `(matplotlib.Figure, matplotlib.Axes)`, optional):
            If you've already created a figure and axes to put the plot in, you 
            can supply them here. Otherwise, `plot_channel_maps` will 
            generate them for you. It will use 
            :code:`visibilities["nrows"][index]` rows and 
            :code:`visibilities["ncols"][index]` columns. Default: `None`
        :attr:`image` (`str`, optional):
            Should the image show the `"data"`, `"model"`, or `"residuals"`.
            Default: `"data"`
        :attr:`contours` (`str`, optional):
            Should the image show the `"data"`, `"model"`, or `"residuals"`.
            Or could also not show contours with `"none"`. Default: `"model"`
        :attr:`model_image` (`str`, optional):
            Should the model image be made by convolving a radiatie transfer
            generated image with an estimate of the beam (`"beam-convolve"`), 
            or by generating model visibilities at the correct baselines of the
            data and then using :code:`interferometry.clean` to generate an 
            image (`"CLEAN"`). Default: `"beam-convolve"`
        :attr:`weighting` (`str`, optional):
            What weighting scheme should the model image use if 
            :code:`model_image="CLEAN"`, `"natural"`, `"robust"`, or 
            `"uniform"`. Default:`"robust"`
        :attr:`robust` (`int`, optional):
            The robust parameter when :code:`weighting="robust"`. Default: 2
        :attr:`maxiter` (`int`, optional):
            The maximum number of CLEAN iterations to perform. Default: 2
        :attr:`threshold` (`float`, optional, Jy):
            The stopping threshold for the CLEAN algorithm. Default: 0.001 Jy
        :attr:`uvtaper` (`float`, optional, lambda):
            FWHM of the Gaussian tapering to apply to the weights when running
            CLEAN. Default: None
        :attr:`image_cmap` (`str`, optional):
            Which colormap to use for plotting the image. Default: `"jet"`
        :attr:`colors_contours` (`str` or `list`-like, optional):
            Colors to use for the contours. If `None`, use the default colormap.
            Default: `None`
        :attr:`levels` (`list` of `float`, optional):
            The flux levels at which to plot contours. If `None`, use 
            :code:`[0.1, 0.3, 0.5, 0.7, 0.9] x image.max()`. Default: `None`
        :attr:`negative_levels` (`list` of `float`, optional):
            A list of negative flux levels at which to plot dashed contours.
            Default: `None`
        :attr:`show_beam` (`bool`, optional):
            Should the plot show the size of the beam? Default: `False`
        :attr:`beamxy` (`tuple`, optional):
            If :code:`show_beam=True`, where should the beam be placed in the
            image, in units of axes fraction. Default: `(0.1,0.1)`
        :attr:`show_colorbar` (`bool`, optional):
            Should the plot show a colorbar for the image? Default: `False` 
        :attr:`cax` (`matplotlib.Axes`, optional):
            Pass an existing Axes class to use for the colorbar. If `None`, 
            :code:`plot_continuum_image` will generate a new one. 
            Default:`None`
        :attr:`colorbar_location` (`str`, optional):
            Should the colorbar be located to the `"right"` of the image or 
            on `"top"` of the image. Default: `'right'`
        :attr:`colorbar_size` (`str`, optional):
            The percent width of the image Axes that should be used for the 
            width of the colorbar Axes. Default: '5%'
        :attr:`colorbar_pad` (`str`, optional):
            How much padding should there be between the image Axes and the 
            colorbar Axes. Default: `0.05`
        :attr:`units` (`str`, optional):
            What units should the colorbar use? Options include `"Jy/beam"`, 
            `"mJy/beam"`, and `"uJy/beam"`. Default: `"Jy/beam"`
        :attr:`show_velocity` (`bool`, optional):
            Label each channel with the velocity of that channel? 
            Default: `True`
        :attr:`vis_marker` (`str`, optional)
            If :code:`plot_vis=True`, the marker to use to plot the visibility
            data. Default: `"o"`
        :attr:`vis_color` (`str`, optional)
            If :code:`plot_vis=True`, the color to use to plot the visibility
            data. Default: `"b"` 
        :attr:`vis_model_color` (`str`, optional)
            If :code:`plot_vis=True`, the color to use to plot the visibility 
            model. Default: `"g"`
        :attr:`show_xlabel` (`bool`, optional)
            Show ticks and axes labels on the x-axis? Default: `True`
        :attr:`show_ylabel` (`bool`, optional)
            Show ticks and axes labels on the y-axis? Default: `True`
        :attr:`skip` (`int`, optional)
            The number of channels to skip between each panel of the plot. 
            Default: `0`
        :attr:`auto_center_velocity` (`bool`, optional)
            Automatically center the velocities within the number of plot 
            panels provided. Default: `False`
        :attr:`v_width` (`float`, optional)
            If :code:`auto_center_velocity=True`, the width of the line that 
            you want to show in the channel maps. :code:`skip` will be 
            automatically determined based on the number of Axes provided. 
            Default: `10.`, kms
        :attr:`fontsize` (`str` or `int`):
            What fontsize to use for labels, ticks, etc. Default: `"medium"`

    Returns:
        :attr:`fig` (`matplotlib.Figure`):
            The matplotlib figure that was used for the plot.
        :attr:`ax` (`matplotlib.Axes`):
            The matplotlib axes that were used for the plot.
        :attr:`cax` (`matplotlib.Axes`, optional):
            The matplotlib axes that were used for the colorbar, if 
            :code:`show_colorbar=True`.
    """

    # Set up the figure if none was provided.

    if fig == None:
        fig, ax = plt.subplots(nrows=visibilities["nrows"][index], \
                ncols=visibilities["ncols"][index], figsize=(10,9))
    else:
        fig, ax = fig

    # Calculate the velocity for each image.

    if plot_vis:
        v = c * (float(visibilities["freq"][index])*1.0e9 - \
                visibilities["data"][index].freq)/ \
                (float(visibilities["freq"][index])*1.0e9)
    else:
        v = c * (float(visibilities["freq"][index])*1.0e9 - \
                visibilities["image"][index].freq)/ \
                (float(visibilities["freq"][index])*1.0e9)

    # Set the ticks.

    ticks = visibilities["image_ticks"][index]

    for i in range(2):
        # First pass through plot the image, second pass through, plot contours.
        if i == 0:
            plot_type = image
        else:
            plot_type = contours

            # If we don't want to plot contours, just skip the second round.
            if plot_type == "none":
                continue

        # Get the correct dataset for what we are plotting (visibilities vs. 
        # image, data vs. model vs. residuals)

        if plot_vis:
            if plot_type == "data":
                vis = visibilities["data1d"][index]
            elif plot_type == "model":
                vis = model.visibilities[visibilities["lam"][index]+"1D"]
            elif plot_type == "residuals":
                residuals = Visibilities(visibilities["data"][index].u, \
                        visibilities["data"][index].v, \
                        visibilities["data"][index].freq, \
                        visibilities["data"][index].real.copy(), \
                        visibilities["data"][index].imag.copy(),\
                        visibilities["data"][index].weights)

                residuals.real -= model.visibilities[visibilities["lam"]\
                        [index]].real
                residuals.imag -= model.visibilities[visibilities["lam"]\
                        [index]].imag

                vis = average(residuals, gridsize=20, radial=True, log=True,\
                    logmin=residuals.uvdist[numpy.nonzero(residuals.uvdist)].\
                    min()*0.95, logmax=residuals.uvdist.max()*1.05, \
                    mode="spectralline")
            elif plot_type == "weights":
                vis = Visibilities(visibilities["data1d"][index].u, \
                        visibilities["data1d"][index].v, \
                        visibilities["data1d"][index].freq, \
                        1./visibilities["data1d"][index].weights**0.5, \
                        1./visibilities["data1d"][index].weights**0.5,\
                        visibilities["data1d"][index].weights.copy())
        else:
            # Get the correct image/contours for plotting.

            if plot_type == "data":
                plot_image = visibilities["image"][index]
            elif plot_type == "model":
                if model_image == "beam-convolve":
                    plot_image = model.images[visibilities["lam"][index]]
                else:
                    model.visibilities[visibilities["lam"][index]].weights = \
                        visibilities["data"][index].weights

                    plot_image = clean(\
                            model.visibilities[visibilities["lam"][index]], \
                            imsize=visibilities["image_npix"][index], \
                            pixel_size=visibilities["image_pixelsize"][index], \
                            weighting=weighting, robust=robust, \
                            convolution="expsinc", mfs=False, \
                            mode="spectralline", maxiter=maxiter, \
                            threshold=threshold, uvtaper=uvtaper)[0]
            elif plot_type == "residuals":
                residuals = Visibilities(visibilities["data"][index].u, \
                        visibilities["data"][index].v, \
                        visibilities["data"][index].freq, \
                        visibilities["data"][index].real.copy(), \
                        visibilities["data"][index].imag.copy(),\
                        visibilities["data"][index].weights)

                residuals.real -= model.visibilities[visibilities["lam"]\
                        [index]].real
                residuals.imag -= model.visibilities[visibilities["lam"]\
                        [index]].imag

                plot_image = clean(residuals, \
                        imsize=visibilities["image_npix"][index], \
                        pixel_size=visibilities["image_pixelsize"][index], \
                        weighting=weighting, robust=robust, \
                        convolution="expsinc", mfs=False, mode="spectralline", \
                        maxiter=0, uvtaper=uvtaper)[0]

            # Get the appropriate scaling for the image.

            if units == "mJy/beam":
                scale = 1000
                colorbar_label = "mJy beam$^{-1}$"
            elif units == "uJy/beam":
                scale = 1e6
                colorbar_label = "$\mu$Jy beam$^{-1}$"
            else:
                scale = 1.
                colorbar_label = "Jy beam$^{-1}$"

            # Set vmin, vmax, and the levels for plotting contours.

            if i == 0:
                if vmin is None:
                    vmin = numpy.nanmin(plot_image.image*scale)
                if vmax is None:
                    vmax = numpy.nanmax(plot_image.image*scale)

            if i == 1:
                if levels is None:
                    levels = numpy.array([0.05, 0.25, 0.45, 0.65, 0.85, 0.95])*\
                            plot_image.image.max()*scale

            # Get the correct range of pixels for making the sub-image.

            if plot_type in ["data","residuals"] or \
                    model_image != "beam-convolve":
                if "x0" in params:
                    x0 = -params["x0"]/visibilities["image_pixelsize"][index]
                else:
                    x0 = -parameters["x0"]["value"]/\
                            visibilities["image_pixelsize"][index]
                if "y0" in params:
                    y0 = params["y0"]/visibilities["image_pixelsize"][index]
                else:
                    y0 = parameters["y0"]["value"]/\
                            visibilities["image_pixelsize"][index]

                xmin = int(round(x0 + visibilities["image_npix"][index]/2 - \
                      visibilities["x0"][index]/\
                      visibilities["image_pixelsize"][index]+ \
                      ticks[0]/visibilities["image_pixelsize"][index]))
                xmax = int(round(x0 + visibilities["image_npix"][index]/2 - \
                      visibilities["x0"][index]/\
                      visibilities["image_pixelsize"][index]+ \
                      ticks[-1]/visibilities["image_pixelsize"][index]))

                ymin = int(round(y0 + visibilities["image_npix"][index]/2 + \
                      visibilities["y0"][index]/\
                      visibilities["image_pixelsize"][index]+ \
                      ticks[0]/visibilities["image_pixelsize"][index]))
                ymax = int(round(y0 + visibilities["image_npix"][index]/2 + \
                      visibilities["y0"][index]/\
                      visibilities["image_pixelsize"][index]+ \
                      ticks[-1]/visibilities["image_pixelsize"][index]))
            else:
                if model_image == "beam-convolve":
                    xmin, xmax = int(round(visibilities["image_npix"][index]/2+\
                            ticks[0]/visibilities["image_pixelsize"][index]+1\
                            )), int(round(visibilities["image_npix"][index]/2 +\
                            ticks[-1]/visibilities["image_pixelsize"][index]+1))
                    ymin, ymax = int(round(visibilities["image_npix"][index]/2+\
                            ticks[0]/visibilities["image_pixelsize"][index]+1\
                            )), int(round(visibilities["image_npix"][index]/2 +\
                            ticks[-1]/visibilities["image_pixelsize"][index]+1))

        # Get the correct starting point, and skip value if auto calculating
        # the velocity range.

        if auto_center_velocity:
            if "v_sys" in params:
                v_start = params["v_sys"] - v_width/2
                v_end = params["v_sys"] + v_width/2
                v_center = params["v_sys"]
            else:
                v_start = parameters["v_sys"] - v_width/2
                v_end = parameters["v_sys"] + v_width/2
                v_center = parameters["v_sys"]

            center = (numpy.abs(v/1e5 - v_center)).argmin()
            start = (numpy.abs(v/1e5 - v_start)).argmin()
            end = (numpy.abs(v/1e5 - v_end)).argmin()

            nchan = end - start + 1
            skip = int(nchan/visibilities["ncols"][index]/\
                    visibilities["nrows"][index] - 0.5)

            if visibilities["ncols"][index]*visibilities["nrows"][index] % 2 \
                    == 0:
                half = int(visibilities["ncols"][index]*\
                        visibilities["nrows"][index] / 2)
            else:
                half = int((visibilities["ncols"][index]*\
                        visibilities["nrows"][index] - 1) / 2)

            start = center - half*(skip + 1)
            end = center + half*(skip + 1)
        else:
            start = visibilities["ind0"][index]

        # Now loop through the channels and plot.

        for k in range(visibilities["nrows"][index]):
            for l in range(visibilities["ncols"][index]):
                ind = (k*visibilities["ncols"][index] + l)*(skip+1) + start

                # Turn off the axis if ind >= nchannels

                if ind >= v.size:
                    print('Index greater than array size, skipping',v.size,ind)
                    ax[k,l].set_axis_off()
                    continue

                # Get a fancy colormap if requested.

                if image_cmap == "BlueToRed":
                    if v[ind]/1e5 < params["v_sys"]:
                        cdict1 = {'red':   ((0.0, 1.0, 1.0),
                                            (1.0, 0.0, 0.0)),
                                  'green': ((0.0, 1.0, 1.0),
                                            (1.0, 0.0, 0.0)),
                                  'blue':  ((0.0, 1.0, 1.0),
                                            (1.0, 1.0, 1.0))}
                        blues = LinearSegmentedColormap('blues', cdict1)
                        cmap = blues
                    else:
                        cdict2 = {'red':   ((0.0, 1.0, 1.0),
                                            (1.0, 1.0, 1.0)),
                                  'green': ((0.0, 1.0, 1.0),
                                            (1.0, 0.0, 0.0)),
                                  'blue':  ((0.0, 1.0, 1.0),
                                            (1.0, 0.0, 0.0))}
                        reds = LinearSegmentedColormap('reds', cdict2)
                        cmap = reds

                    scalefn = lambda x: (numpy.arctan(x*7-4)+1.3) / \
                            (numpy.arctan(1*7-4)+1.3)
                else:
                    cmap = image_cmap
                    scalefn = lambda x: 1.

                # On the first loop through, plot the image, and on the second
                # the contours.

                if i == 0:
                    # Plot the image.

                    if plot_vis:
                        ax[k,l].errorbar(vis.uvdist/1000, vis.amp[:,ind], \
                                yerr=1./vis.weights[:,ind]**0.5, fmt=\
                                vis_color+vis_marker)
                    else:
                        implot = ax[k,l].imshow(plot_image.image[ymin:ymax,\
                                xmin:xmax,ind,0]*scale*scalefn(abs(v[ind]/1e5 -\
                                params["v_sys"])), origin="lower", \
                                interpolation="nearest", vmin=vmin, vmax=vmax, \
                                cmap=cmap)

                    # Add the velocity to the map.

                    if show_velocity:
                        if k == 0 and l == 0:
                            txt = ax[k,l].annotate(r"$v=%{0:s}$ km s$^{{-1}}$".\
                                    format(visibilities["fmt"][index]) % \
                                    (v[ind]/1e5), xy=(0.95,0.85), \
                                    xycoords='axes fraction', \
                                    horizontalalignment="right", \
                                    fontsize=fontsize)
                        else:
                            txt = ax[k,l].annotate(r"$%{0:s}$ km s$^{{-1}}$".\
                                    format(visibilities["fmt"][index]) % \
                                    (v[ind]/1e5), xy=(0.95,0.85), \
                                    xycoords='axes fraction', \
                                    horizontalalignment="right", \
                                    fontsize=fontsize)

                    # Fix the axes labels.

                    if plot_vis:
                        if show_xlabel:
                            ax[-1,l].set_xlabel("Baseline [k$\lambda$]", \
                                    fontsize=fontsize)

                        if plot_type == "residuals":
                            ax[k,l].set_ylim(-vis.amp.max()*1.1, \
                                    vis.amp.max()*1.1)
                        else:
                            ax[k,l].set_ylim(-0.5, vis.amp.max()*1.1)

                        ax[k,l].set_xscale("log", nonpositive='clip')

                        # Turn off tick labels if we aren't in the correct
                        # row/column.
                        if l > 0:
                            ax[k,l].set_yticklabels([])
                        if k < visibilities["nrows"][index]-1:
                            ax[k,l].set_xticklabels([])
                    else:
                        # Set the ticks based on the config file.

                        transformx = ticker.FuncFormatter(Transform(xmin, xmax,\
                                visibilities["image_pixelsize"][index],'%.1f"'))
                        transformy = ticker.FuncFormatter(Transform(ymin, ymax,\
                                visibilities["image_pixelsize"][index],'%.1f"'))

                        ax[k,l].set_xticks((xmin + xmax)/2+\
                                ticks[1:-1]/visibilities["image_pixelsize"]\
                                [index]-xmin)
                        ax[k,l].set_yticks((ymin + ymax)/2+\
                                ticks[1:-1]/visibilities["image_pixelsize"]\
                                [index]-ymin)

                        ax[k,l].get_xaxis().set_major_formatter(transformx)
                        ax[k,l].get_yaxis().set_major_formatter(transformy)

                        # Adjust the tick labels.

                        ax[k,l].tick_params(labelsize=fontsize)

                        # Add a label to the x-axis.

                        if show_xlabel:
                            ax[-1,l].set_xlabel("$\Delta$RA", fontsize=fontsize)

                        # Show the size of the beam.

                        if show_beam:
                            bmaj = visibilities["image"][index].header["BMAJ"]/\
                                    abs(visibilities["image"][index].\
                                    header["CDELT1"])
                            bmin = visibilities["image"][index].header["BMIN"]/\
                                    abs(visibilities["image"][index].\
                                    header["CDELT1"])
                            bpa = visibilities["image"][index].header["BPA"]

                            xy = ((xmax-xmin)*beamxy[0], (ymax-ymin)*beamxy[1])

                            ax[k,l].add_artist(patches.Ellipse(xy=xy, \
                                    width=bmaj, height=bmin, angle=(bpa+90), \
                                    facecolor="white", edgecolor="black"))

                        ax[k,l].set_adjustable('box')

                    if plot_vis:
                        if show_ylabel:
                            ax[k,0].set_ylabel("Amplitude [Jy]", \
                                    fontsize=fontsize)
                    else:
                        if show_ylabel:
                            ax[k,0].set_ylabel("$\Delta$Dec", fontsize=fontsize)

                # Plot the contours.

                elif i == 1:
                    if plot_vis:
                        ax[k,l].plot(vis.uvdist/1000, vis.amp[:,ind], \
                                vis_model_color+"-")
                    else:
                        if len(numpy.where(plot_image.image[ymin:ymax,\
                                xmin:xmax,ind,0]*scale > levels.min())[0]) > 0:
                            ax[k,l].contour(plot_image.image[ymin:ymax,\
                                    xmin:xmax,ind,0]*scale, levels=levels, \
                                    colors=contours_colors)

                        # Plot the negative contours, if requested.
                        if negative_levels is not None:
                            if len(numpy.where(plot_image.image[ymin:ymax,\
                                    xmin:xmax,ind,0]*scale < \
                                    negative_levels.max())[0]) > 0:
                                ax[k,l].contour(plot_image.image[ymin:ymax,\
                                        xmin:xmax,ind,0]*scale, \
                                        levels=negative_levels, \
                                        linestyles="--", colors=contours_colors)

        # Add a colorbar to the image.

        if not plot_vis and show_colorbar:
            # If no cax were provided, create them based on the options.

            user_specified_cbar = True

            if cax == None:
                ax_bbox = ax[0,visibilities["ncols"][index]-1].get_position()

                if colorbar_location == 'top':
                    cax = fig.add_axes([ax_bbox.x0, ax_bbox.y1+colorbar_pad, \
                            ax_bbox.width, float(colorbar_size[0:-1])/100*\
                            ax_bbox.height])

                    colorbar_orientation = "horizontal"
                else:
                    cax = fig.add_axes([ax_bbox.x1+colorbar_pad, ax_bbox.y0, \
                            float(colorbar_size[0:-1])/100*ax_bbox.width, \
                            ax_bbox.height])

                    colorbar_orientation = "vertical"

                user_specified_cbar = False

            # Now plot the colorbar.

            cbar = plt.colorbar(implot, cax=cax, \
                    orientation=colorbar_orientation)

            # And make some adustments.

            if not user_specified_cbar:
                if colorbar_location == 'top':
                    cax.xaxis.set_ticks_position('top')
                    cax.xaxis.set_label_position('top')

            cbar.set_label(colorbar_label, size=fontsize)

            cax.tick_params(axis="both", which="major", labelsize=fontsize)

    # Return the figure and axes.

    if show_colorbar:
        return fig, ax, cax
    else:
        return fig, ax

# Define a useful class for plotting.

class Transform:
    def __init__(self, xmin, xmax, dx, fmt):
        self.xmin = xmin
        self.xmax = xmax
        self.dx = dx
        self.fmt = fmt

    def __call__(self, x, p):
        return self.fmt% ((x-(self.xmax-self.xmin+1)/2)*self.dx)



from scipy.signal import fftconvolve
from scipy.optimize import leastsq
from ..imaging import Image
from .invert import invert
import numpy

def clean(data, imsize=256, pixel_size=0.25, convolution="pillbox", mfs=False,\
        weighting="natural", robust=2, npixels=0, centering=None, \
        mode='continuum', gain=0.1, maxiter=1000, threshold=0.001, \
        uvtaper=None):

    # First make the image.

    image = invert(data, imsize=imsize, pixel_size=pixel_size, \
            convolution=convolution, mfs=mfs, weighting=weighting, \
            robust=robust, npixels=npixels, centering=centering, mode=mode, \
            uvtaper=uvtaper)

    # Now, also make an image of the beam.

    beam = invert(data, imsize=2*imsize, pixel_size=pixel_size, \
            convolution=convolution, mfs=mfs, weighting=weighting, \
            robust=robust, npixels=npixels, centering=centering, mode=mode, \
            beam=True)

    # Now start the clean-ing by defining some variables.

    dirty = image.image[:,:,:,0]
    dirty_beam = beam.image[:,:,:,0]
    wherezero = dirty == 0
    nonzero = dirty != 0
    
    model = numpy.zeros(dirty.shape)

    # Calculate the size of the beam.

    clean_beam = numpy.zeros(dirty_beam.shape)

    ny, nx, nfreq = dirty_beam.shape
    x, y = numpy.meshgrid(numpy.arange(nx) - nx/2 + 1, numpy.arange(ny) - ny/2)

    for i in range(nfreq):
        fitfunc = lambda p, x, y: numpy.exp(-(x * numpy.cos(p[2]) - \
                y * numpy.sin(p[2]))**2 / (2*p[0]**2) - (x * numpy.sin(p[2]) + \
                y * numpy.cos(p[2]))**2 / (2*p[1]**2))
        errfunc = lambda p, x, y, z, w: numpy.ravel((fitfunc(p, x, y) - z) * w)
        p0 = [0.5,0.5,0.]

        weights = numpy.abs(dirty_beam[:,:,i])*(dirty_beam[:,:,i] > 0.4)

        p, success = leastsq(errfunc, p0, args=(x,y,dirty_beam[:,:,i],weights))

        clean_beam[:,:,i] = fitfunc(p, x, y)
    
    # Generate a mask.

    mask = numpy.ones(dirty.shape)
    """
    TODO: Add in auto-masking.
    """

    # Now loop through and subtract off the beam.

    n = 0
    stop = False
    while n < maxiter and not stop:
        # Determine the location of the maximum value inside the mask.

        maxval = dirty*mask == (dirty*mask).max()

        # Add that value to the model (with some gain).
        
        model[maxval] = model[maxval] + dirty[maxval]*gain

        # Also subtract off that value from the image.
        
        subtract = numpy.zeros(dirty.shape)
        subtract[maxval] = dirty[maxval]*gain
        
        for i in range(nfreq):
            dirty[:,:,i] = dirty[:,:,i] - fftconvolve(subtract[:,:,i], \
                    dirty_beam[:,:,i], mode='same')
        dirty[wherezero] = 0.
        
        # Do we stop here?

        stop = (dirty*mask).max() < threshold

        n = n + 1

    # Generate a clean beam and convolve with the model to make a cleaned image.

    clean_image = numpy.zeros(dirty.shape)
    for i in range(nfreq):
        clean_image[:,:,i] = fftconvolve(model[:,:,i],clean_beam[:,:,i], \
                mode='same')+dirty[:,:,i]
    
    model = Image(model.reshape((model.shape[0],model.shape[1],\
            model.shape[2],1)), freq=data.freq)
    residuals = Image(dirty.reshape((dirty.shape[0],dirty.shape[1],\
            dirty.shape[2],1)), freq=data.freq)
    clean_beam = Image(clean_beam.reshape((dirty_beam.shape[0],\
            dirty_beam.shape[1],dirty_beam.shape[2],1)), freq=data.freq)
    clean_image = Image(clean_image.reshape(model.image.shape), \
            freq=data.freq)
    
    return clean_image, residuals, beam, model



#!/usr/bin/env python3

from pdspy.dust import *
import numpy

water_ice = Dust()
water_ice.set_optical_constants_from_henn("optical_constants/water_ice.txt")
water_ice.set_density(0.92)

#1/3
graphite_parallel = Dust()
graphite_parallel.set_optical_constants_from_draine("optical_constants/graphite_parallel_0.01.txt")
graphite_parallel.set_density(2.24)
graphite_parallel.calculate_optical_constants_on_wavelength_grid(water_ice.lam)

#2/3
graphite_perpendicular = Dust()
graphite_perpendicular.set_optical_constants_from_draine("optical_constants/graphite_perpendicular_0.01.txt")
graphite_perpendicular.set_density(2.24)
graphite_perpendicular.calculate_optical_constants_on_wavelength_grid(\
        water_ice.lam)

silicates = Dust()
silicates.set_optical_constants_from_draine("optical_constants/astronomical_silicates.txt")
silicates.set_density(3.3)
silicates.calculate_optical_constants_on_wavelength_grid(water_ice.lam)

species = [silicates,graphite_parallel,graphite_perpendicular]
abundances = numpy.array([0.65,0.35*1./3,0.35*2./3])
print(abundances)

dust = mix_dust(species, abundances)

amin = 0.005e-4
amax = 2.000e-1
pl = 3.5

dust.calculate_size_distribution_opacity(amin, amax, pl, nang=1, \
        coat_volume_fraction=0.0)

dust.write('draine_2mm.hdf5')



#!/usr/bin/env python3

from pdspy.dust import Dust
import matplotlib
import matplotlib.pyplot as plt

# List of the files to be plotted.

species_list = ["draine_1um.hdf5", "draine_10um.hdf5", "draine_100um.hdf5", \
        "draine_1mm.hdf5", "draine_3mm.hdf5", "draine_1cm.hdf5"]

# Change a few of the parameters to make the plot look nice.

matplotlib.rcParams["font.family"] = "serif"
matplotlib.rcParams["font.size"] = 14
matplotlib.rcParams["text.usetex"] = "True"
matplotlib.rcParams["text.latex.preamble"] = r"\usepackage{upgreek}"
matplotlib.rcParams["legend.fontsize"] = 14

# Start the plotting.

fig, ax = plt.subplots(nrows=2, ncols=2)

for species in species_list:
    dust = Dust()
    dust.set_properties_from_file(species)

    # Make a label for each line.

    size = species.split('_')[1].split('.')[0]
    if (len(size.split('um')) == 2):
        label = r"$a_{max} = %s$ $\upmu$m" % size.split('um')[0]
    elif (len(size.split('mm')) == 2):
        label = "$a_{max} = %s$ mm" % size.split('mm')[0]
    elif (len(size.split('cm')) == 2):
        label = "$a_{max} = %s$ cm" % size.split('cm')[0]

    # Plot the opacities.

    ax[0,0].loglog(dust.lam*1e4, dust.kabs, label=label)
    ax[0,1].loglog(dust.lam*1e4, dust.ksca)
    ax[1,0].loglog(dust.lam*1e4, dust.kext)
    ax[1,1].semilogx(dust.lam*1e4, dust.albedo)

ax[0,0].set_xlabel(r"$\lambda$ [$\upmu$m]")
ax[0,0].set_ylabel("$\kappa_{abs}$ [cm$^2$ g$^{-1}$]")
ax[0,1].set_xlabel(r"$\lambda$ [$\upmu$m]")
ax[0,1].set_ylabel("$\kappa_{sca}$ [cm$^2$ g$^{-1}$]")
ax[1,0].set_xlabel(r"$\lambda$ [$\upmu$m]")
ax[1,0].set_ylabel("$\kappa_{ext}$ [cm$^2$ g$^{-1}$]")
ax[1,1].set_xlabel(r"$\lambda$ [$\upmu$m]")
ax[1,1].set_ylabel("Albedo")

ax[0,0].legend(loc="lower left")

ax[0,0].axis([1e-1,1e5,1e-4,1e5])
ax[0,1].axis([1e-1,1e5,1e-4,1e5])
ax[1,0].axis([1e-1,1e5,1e-4,1e5])
ax[1,1].axis([1e-1,1e5,0,1])

fig.set_size_inches((10,10))
fig.subplots_adjust(wspace=0.25, left=0.08, top=0.97, right=0.98, bottom=0.06)
fig.savefig("draine_plot.pdf")



import numpy
from .average import average
from .grid import grid
from .clean import clean
from ..imaging import Image
from ..constants.physics import c
from ..constants.math import pi
from ..constants.astronomy import arcsec
from scipy.fftpack import ifft2, fftshift, ifftshift
from scipy.special import jn

def invert(data, imsize=256, pixel_size=0.25, weights="natural", \
        convolution="pillbox", mfs=False, spectral=False, restfreq=None, \
        start_velocity=None, beam=False, nchannels=1, clean=False, \
        maxiter=1000, threshold=0.001, box=None):
    
    kms = 1.0e5
    
    if (data.freq.size > 1) and (mfs == False) and (spectral == False):
        averaged_data = average(data,only_freq=True)
    else:
        averaged_data = data.copy()

    if beam:
        averaged_data.real[:,:] = 1.
        averaged_data.imag[:,:] = 0.
    
    maxD = 0.

    if spectral:
        velocity = c * (restfreq - averaged_data.freq) / restfreq
        startchannel = numpy.where(abs(velocity - start_velocity*kms) == \
                abs(velocity-start_velocity*kms).min())[0][0]
    else:
        startchannel = 0

    for i in range(startchannel, startchannel + nchannels):
        baselines = numpy.unique(averaged_data.baseline)

        im = numpy.zeros((imsize, imsize))
        response = numpy.zeros((imsize, imsize))

        for j in baselines:
            baseline_data = averaged_data.get_baselines(j)
            if beam:
                weight = baseline_data.weights[baseline_data.real != 0.].sum()
            else:
                weight = baseline_data.weights[(baseline_data.real != 0.) \
                        & (baseline_data.imag != 0.)].sum()

            binsize = 1.0 / (pixel_size * imsize * arcsec)
            gridded_data = grid(baseline_data, gridsize=imsize, \
                    binsize=binsize, convolution=convolution, mfs=mfs, \
                    channel=i)
            
            real = gridded_data.real.reshape((imsize, imsize))
            imag = gridded_data.imag.reshape((imsize, imsize))
    
            if weights == "natural":
                weights = gridded_data.weights.reshape((imsize, imsize)).T
            elif weights == "uniform":
                weights = gridded_data.weights.reshape((imsize, imsize)).T
                weights[weights > 0] = 1.0

            comp = real + 1j*imag
    
            x = linspace(-imsize/2, imsize/2-1, imsize)*pixel_size
            y = linspace(-imsize/2, imsize/2-1, imsize)*pixel_size
            xarray, yarray = meshgrid(x, y)
            rarray = sqrt(xarray**2+yarray**2)
    
            if convolution == "pillbox":
                convolve_func = pillbox
            elif convolution == "expsinc":
                convolve_func = exp_sinc
    
            u = gridded_data.u.reshape((imsize,imsize))
            v = gridded_data.v.reshape((imsize,imsize))
            
            convolve = convolve_func(u,v,binsize,binsize)
            convolve = fftshift(fft2(fftshift(convolve))).real

            D1 = float(j.split("-")[0])/(c/data.freq[i])*arcsec
            D2 = float(j.split("-")[1])/(c/data.freq[i])*arcsec

            if D1 > maxD:
                maxD = D1
            elif D2 > maxD:
                maxD = D2

            nonzero = rarray != 0.
            responsivity = numpy.zeros(xarray.shape)
            responsivity[nonzero] = 4*jn(1,pi*rarray[nonzero]*D1)* \
                    jn(1,pi*rarray[nonzero]*D2)/((pi*rarray[nonzero])**2*D1*D2)
            responsivity[rarray == 0.] = 1.

            if clean:
                image = fftshift(fft2(fftshift(comp))).real/convolve
                beam = fftshift(fft2(fftshift(weights))).real/convolve

                cleanim = clean(Image(image.reshape((imsize,imsize,1))), \
                        Image(beam.reshape((imsize,imsize,1))), \
                        maxiter=maxiter, threshold=threshold, \
                        box=box).image[:,:,0]

                im += responsivity * cleanim
            else:
                im += responsivity * fftshift(fft2(fftshift(comp))).real/ \
                        convolve*weight

            response += responsivity*weight

        im[rarray > 1.0/maxD] = 0.
        im = im/response

        if i == startchannel:
            image = im.reshape((imsize,imsize,1))
        else:
            image = concatenate((image,im),axis=2)

    if spectral:
        return Image(image, \
                velocity=velocity[startchannel:startchannel+nchannels])
    else:
        return Image(image)

def pillbox(u,v,delta_u,delta_v):
    
    arr = zeros(u.shape)
    arr[u.shape[0]/2.,u.shape[1]/2.] = 1.0
    
    return arr

def exp_sinc(u,v,delta_u,delta_v):
    
    alpha1 = 1.55
    alpha2 = 2.52
    
    return sinc(u/(alpha1*delta_u))*exp(-1*(u/(alpha2*delta_u))**2)* \
           sinc(v/(alpha1*delta_u))*exp(-1*(v/(alpha2*delta_v))**2)

def spheroidal(u,v,delta_u,delta_v):
    
    alpha = 0.0
    m = 6.0
    
    return abs(1-(2*u/(m*delta_u))**2)**alpha



from numpy import array, arange, exp
from numpy.random import uniform
from .change_params import change_params
from .ml2d import ml2d

def mcmc2d(x, y, z, sigma_z, params, sigma_params, model, args={}, \
           nsteps=1e5, change_param=None, limits=None):

    MLold, chisq_old = ml2d(x, y, z, sigma_z, params, model, args=args, \
            limits=limits)

    accepted_params = []

    for i in arange(nsteps):
        new_params = change_params(params, sigma_params, \
                change_param=change_param)

        MLnew, chisq_new = ml2d(x, y, z, sigma_z, new_params, model, \
                args=args, limits=limits)

        if chisq_new < chisq_old:
            accepted_params.append(params)
            params = new_params
            MLold = MLnew
            chisq_old = chisq_new
        else:
            if uniform(low=0., high=1., size=1)[0] <= \
                    exp(0.5*(chisq_old-chisq_new)):
                accepted_params.append(params)
                params = new_params
                MLold = MLnew
                chisq_old = chisq_new

    return array(accepted_params)



from ..constants.astronomy import arcsec
from ..constants.physics import c
from .libimaging import Image
from astropy.utils.exceptions import AstropyWarning
import astropy.io.fits as fits
import astropy.wcs as wcs
import warnings
import numpy

def readpvfits(filename):
    
    # Open the fits file.

    data = fits.open(filename)

    # Figure out the dimensions of each axis and create an array to put the data
    # into, and put the data into that array.

    npol, nv, nx = data[0].data.shape
    
    image = numpy.zeros((1,nx,nv,npol))

    for i in range(npol):
        for j in range(nv):
            image[0,:,j,i] = data[0].data[i,j,:]

    # Read in the x and y coordinate information, including the WCS info if it
    # is available.

    header = data[0].header

    #x, y = numpy.meshgrid(numpy.linspace(0,nx-1,nx), numpy.linspace(0,ny-1,ny))
    x, y = None, None

    x0 = data[0].header["CRVAL1"]
    dx = data[0].header["CDELT1"]
    nx0 = data[0].header["CRPIX1"]

    x = (numpy.arange(nx) - (nx0-1))*dx + x0

    nu0 = data[0].header["CRVAL2"]
    dnu = data[0].header["CDELT2"]
    n0 = data[0].header["CRPIX2"]
    freq = (numpy.arange(nv)-(n0-1))*dnu + nu0
    restfreq = data[0].header["RESTFRQ"]

    velocity = c * (restfreq - freq) / restfreq

    data.close()
 
    return Image(image, x=x, y=y, header=header, wcs=None, velocity=velocity, \
            freq=freq)



import pdspy.interferometry as uv
import matplotlib.pyplot as plt
import numpy

def plot_1D_visibilities(visibilities, model, parameters, params, index=0, \
        fig=None, plot_disk=False, color="k", markersize=8, linewidth=1, \
        line_color="g", disk_only_color="gray", fontsize="medium"):
    r"""
    Plot the 1D azimuthally averaged visibility data along with the specified model.

    Args:
        :attr:`visibilities` (`dict`):
            Dictionary containing the visibility data, typically as loaded by :code:`utils.load_config` and :code:`utils.load_data`.
        :attr:`model` (`modeling.Model`):
            The radiative transfer model that you would like to plot the visibilities of. Typically this is the output of modeling.run_disk_model.
        :attr:`parameters` (`dict`):
            The parameters dictionary in the config module as loaded in by :code:`utils.load_config`
        :attr:`params` (`dict`):
            The parameters of the model, typically as a dictionary mapping parameter keys from the :code:`parameters` dictionary to their values.
        :attr:`index` (`int`, optional):
            The visibilities dictionary typically contains a list of datasets. `index` indicates which one to plot.
        :attr:`fig` (`tuple`, `(matplotlib.Figure, matplotlib.Axes)`, optional):
            If you've already created a figure and axes to put the plot in, you can supply them here. Otherwise, `plot_1D_visibilities` will generate them for you. Default: `None`
        :attr:`plot_disk` (`bool`, optional):
            Should :code:`plot_1D_visibilities` show the disk-only contribution to the model? Default: `False`
        :attr:`color` (str, optional):
            The color to use for plotting the visibility data. Default: `"k"`
        :attr:`markersize` (`int`, optional):
            The size of the markers to use for plotting the visibility data. Default: `8`
        :attr:`linewidth` (int, optional):
            What linewidth to use for plotting the model. Default: 1
        :attr:`line_color` (str, optional):
            The color to use for plotting the model visibilities. Default: `"g"`
        :attr:`disk_only_color` (str, optional):
            The color to use for plotting the disk-only model visibilities. Default: `"gray"`
        :attr:`fontsize` (`str` or `int`):
            What fontsize to use for labels, ticks, etc. Default: `"medium"`

    Returns:
        :attr:`fig` (`matplotlib.Figure`):
            The matplotlib figure that was used for the plot.
        :attr:`ax` (`matplotlib.Axes`):
            The matplotlib axes that were used for the plot.
    """


    # Generate a figure and axes if not provided.

    if fig == None:
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(4.5,4))
    else:
        fig, ax = fig

    # Average the high resolution model radially.

    m1d = uv.average(model.visibilities[visibilities["lam"][index]+"_high"], \
            gridsize=10000, binsize=3500, radial=True)

    if plot_disk:
        m1d_disk = uv.average(model.visibilities[visibilities["lam"][index]+\
                "_disk"], gridsize=10000, binsize=3500, radial=True)

    # Calculate the error properly.

    real_samples = visibilities["data1d"][index].real + numpy.random.normal(\
            0, 1, (visibilities["data1d"][index].real.size, 1000)) * \
            1./visibilities["data1d"][index].weights**0.5

    imag_samples = visibilities["data1d"][index].imag + numpy.random.normal(\
            0, 1, (visibilities["data1d"][index].imag.size, 1000)) * \
            1./visibilities["data1d"][index].weights**0.5

    amp_samples = (real_samples**2 + imag_samples**2)**0.5

    amp_unc = (numpy.percentile(amp_samples, [50], axis=1) - \
            numpy.percentile(amp_samples, [16, 84], axis=1)) * \
            numpy.array([1,-1])[:,None]

    # Plot the visibilities.

    ax.errorbar(visibilities["data1d"][index].uvdist/1000, \
            visibilities["data1d"][index].amp*1000, \
            yerr=amp_unc*1000, \
            fmt="o", markersize=markersize, markeredgecolor=color, \
            markerfacecolor=color, ecolor=color)

    # Plot the best fit model

    if plot_disk:
        ax.plot(m1d_disk.uvdist/1000, m1d_disk.amp*1000, "--", \
                color=disk_only_color, linewidth=linewidth)

    ax.plot(m1d.uvdist/1000, m1d.amp*1000, "-", color=line_color, \
            linewidth=linewidth)

    # Adjust the plot and add axes labels.

    ax.axis([1,visibilities["data1d"][index].uvdist.max()/1000*3,0,\
            visibilities["data1d"][index].amp.max()*1.1*1000])

    ax.set_xscale("log", nonpositive='clip')

    ax.set_xlabel("Baseline [k$\lambda$]", fontsize=fontsize)
    ax.set_ylabel("Amplitude [mJy]", fontsize=fontsize)

    # Return the figure and axes.

    return fig, ax



from pdspy.constants.physics import m_p, c, h
import numpy
import h5py

class Gas:

    def set_properties_from_lambda(self, filename):
        f = open(filename)

        for i in range(3):
            f.readline()

        self.mass = float(f.readline())

        f.readline()
        nlev = int(f.readline())
        f.readline()

        self.J = numpy.empty(nlev, dtype="<U6")
        self.E = numpy.empty(nlev, dtype=float)
        self.g = numpy.empty(nlev, dtype=float)
        for i in range(nlev):
            temp, self.E[i], self.g[i], self.J[i] = tuple(f.readline().split())

        f.readline()
        ntrans = int(f.readline())
        f.readline()

        self.J_u = numpy.empty(ntrans, dtype=int)
        self.J_l = numpy.empty(ntrans, dtype=int)
        self.A_ul = numpy.empty(ntrans, dtype=float)
        self.nu = numpy.empty(ntrans, dtype=float)
        self.E_u = numpy.empty(ntrans, dtype=float)
        for i in range(ntrans):
            temp, self.J_u[i], self.J_l[i], self.A_ul[i], self.nu[i], \
                    self.E_u[i] = tuple(f.readline().split())

        self.nu *= 1.0e9
        self.B_ul = c**2 * self.A_ul / (2*h*self.nu**3)

        f.readline()
        npartners = int(f.readline())

        self.partners = []
        self.temp = []
        self.J_u_coll = []
        self.J_l_coll = []
        self.gamma = []
        for i in range(npartners):
            f.readline()
            self.partners.append(f.readline())
            f.readline()
            ncolltrans = int(f.readline())
            f.readline()
            ncolltemps = int(f.readline())
            f.readline()
            self.temp.append(numpy.array(f.readline().split(), dtype=float))
            f.readline()

            self.J_u_coll.append(numpy.empty(ncolltrans, dtype=int))
            self.J_l_coll.append(numpy.empty(ncolltrans, dtype=int))
            self.gamma.append(numpy.empty((ncolltrans,ncolltemps), \
                    dtype=float))

            for j in range(ncolltrans):
                temp, self.J_u_coll[i][j], self.J_l_coll[i][j], temp2 = \
                        tuple(f.readline().split(None,3))
                self.gamma[i][j,:] = numpy.array(temp2.split())

        f.close()

    def set_properties_from_file(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "r")
        else:
            f = usefile

        self.mass = f['mass'][()]

        self.J = f["J"][...]
        self.E = f['E'][...]
        self.g = f['g'][...]

        self.J_u = f['J_u'][...]
        self.J_l = f['J_l'][...]
        self.A_ul = f['A_ul'][...]
        self.nu = f['nu'][...]
        self.E_u = f['E_u'][...]

        self.B_ul = c**2 * self.A_ul / (2*h*self.nu**3)

        self.partners = []
        self.temp = []
        self.J_u_coll = []
        self.J_l_coll = []
        self.gamma = []
        for name in f["CollisionalPartners"]:
            self.partners.append(name)
            self.temp.append(f["CollisionalPartners"][name] \
                    ["Temperature"][...])
            self.J_u_coll.append(f["CollisionalPartners"][name] \
                    ["J_u_coll"][...])
            self.J_l_coll.append(f["CollisionalPartners"][name] \
                    ["J_l_coll"][...])
            self.gamma.append(f["CollisionalPartners"][name]["Gamma"][...])

        if (usefile == None):
            f.close()

    def write(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "w")
        else:
            f = usefile

        f['mass'] = self.mass

        J_dset = f.create_dataset("J", (self.J.size,), dtype=h5py.special_dtype(vlen=str))
        J_dset[...] = self.J
        E_dset = f.create_dataset("E", (self.E.size,), dtype='f')
        E_dset[...] = self.E
        g_dset = f.create_dataset("g", (self.g.size,), dtype='f')
        g_dset[...] = self.g

        J_u_dset = f.create_dataset("J_u", (self.J_u.size,), dtype='f')
        J_u_dset[...] = self.J_u
        J_l_dset = f.create_dataset("J_l", (self.J_l.size,), dtype='f')
        J_l_dset[...] = self.J_l
        A_ul_dset = f.create_dataset("A_ul", (self.A_ul.size,), dtype='f')
        A_ul_dset[...] = self.A_ul
        nu_dset = f.create_dataset("nu", (self.nu.size,), dtype='f')
        nu_dset[...] = self.nu
        E_u_dset = f.create_dataset("E_u", (self.E_u.size,), dtype='f')
        E_u_dset[...] = self.E_u

        collisions = f.create_group("CollisionalPartners")
        partners = []
        temp = []
        J_u_coll = []
        J_l_coll = []
        gamma = []
        for i in range(len(self.partners)):
            partners.append(collisions.create_group("{0:s}".format( \
                    self.partners[i])))

            temp.append(partners[i].create_dataset("Temperature", \
                    (self.temp[i].size,), dtype='f'))
            temp[i][...] = self.temp[i]

            J_u_coll.append(partners[i].create_dataset("J_u_coll", \
                    (self.J_u_coll[i].size,), dtype='f'))
            J_u_coll[i][...] = self.J_u_coll[i]

            J_l_coll.append(partners[i].create_dataset("J_l_coll", \
                    (self.J_l_coll[i].size,), dtype='f'))
            J_l_coll[i][...] = self.J_l_coll[i]

            gamma.append(partners[i].create_dataset("Gamma", \
                    self.gamma[i].shape, dtype='f'))
            gamma[i][...] = self.gamma[i]

        if (usefile == None):
            f.close()



import scipy.integrate
import numpy
from ..constants.physics import G, m_p, k, m_p
from ..constants.astronomy import AU, M_sun
from .Disk import Disk

class DartoisDisk(Disk):
    def log_gas_density_high(self, mstar=0.5):
        # Set up the high resolution grid.

        r = numpy.logspace(numpy.log10(self.rmin), numpy.log10(5*self.rmax), \
                1000)
        z = numpy.hstack(([0], numpy.logspace(-3., numpy.log10(5*self.rmax), \
                1000)))
        phi = numpy.array([0.,2*numpy.pi])

        # Calculate the temperature structure.

        T = self.temperature(r, phi, z, coordsys="cylindrical")[:,0,:]

        # Now make the high resolution coordinates into 3D arrays.

        r, z = numpy.meshgrid(r, z, indexing='ij')

        # Now calculate the derivative of T.

        dlogT = numpy.gradient(numpy.log(T), z[0,:]*AU, axis=1)

        # Calculate the balance of gravity and pressure

        mu = 2.37
        cs = (k*T/(mu*m_p))**0.5
        M = mstar*M_sun

        gravity = 1./cs**2 * G*M*z*AU / ((r*AU)**2 + (z*AU)**2)**1.5

        # Now vertically integrate to get -logDens.

        logDens = -scipy.integrate.cumtrapz(dlogT+gravity, x=z*AU, initial=0., \
                axis=1)

        # Get the proper normalization.

        norm = 0.5*100*self.surface_density(r[:,0]) / scipy.integrate.trapz(\
                numpy.exp(logDens), x=z*AU, axis=1)

        norm_arr = numpy.zeros(r.shape)
        for i in range(r.shape[1]):
            norm_arr[:,i] = norm

        # Finally, calculate the scaled density.

        with numpy.errstate(divide="ignore"):
            logDens += numpy.where(norm_arr > 0, numpy.log(norm_arr), -300.)

        return r, z, logDens

    def log_number_density_high(self, mstar=0.5, gas=0):
        # Get the high resolution gas density in cylindrical coordinates.

        r, z, logrho_gas = self.log_gas_density_high(mstar=mstar)

        logn_H2 = logrho_gas + numpy.log(0.8 / (2.37*m_p))

        if gas < 0:
            return r, z, logn_H2
        else:
            # Take account of the abundance.

            logn = logn_H2 + numpy.log(self.abundance[gas])

            # Account for freezeout as well.

            T = self.temperature(r[:,0], numpy.array([0.,2*numpy.pi]), z[0,:], \
                    coordsys="cylindrical")[:,0,:]

            frozen = T < self.freezeout[gas]

            # Account for photodissociation.

            n_H2 = numpy.exp(logn_H2)

            cumn = -scipy.integrate.cumtrapz(n_H2[:,::-1], x=z[:,::-1]*AU, \
                    axis=1, initial=0.)[:,::-1]

            dissociated = cumn < 0.79 * 1.59e21 / 0.706

            # Now actually reduce the density.

            logn[numpy.logical_or(frozen, dissociated)] += numpy.log(1.0e-8)

            return r, z, logn

    def log_gas_pressure_high(self, mstar=0.5):
        # Get the number density and temperature.

        r, z, logn = self.log_number_density_high(mstar=mstar, gas=-1)

        T = self.temperature(r[:,0], numpy.array([0.,2*numpy.pi]), z[0,:], \
                coordsys="cylindrical")[:,0,:]

        logT = numpy.log(T)

        # And calculate the pressure.

        logP = logn + logT + numpy.log(k)

        return r, z, logP


    def gas_density(self, x1, x2, x3, coordsys="spherical", mstar=0.5):
        ##### Set up the coordinates

        if coordsys == "spherical":
            rt, tt, pp = numpy.meshgrid(x1*AU, x2, x3, indexing='ij')

            rr = rt*numpy.sin(tt)
            zz = rt*numpy.cos(tt)
        elif coordsys == "cartesian":
            xx, yy, zz = numpy.meshgrid(x1*AU, x2*AU, x3*AU, indexing='ij')

            rr = (xx**2 + yy**2)**0.5
        elif coordsys == "cylindrical":
            rr, pp, zz = numpy.meshgrid(x1*AU, x2, x3*AU, indexing='ij')

        # Get the high resolution log of the gas density.

        r, z, logDens = self.log_gas_density_high(mstar=mstar)

        # Now, interpolate that density onto the actual grid of interest.

        f = scipy.interpolate.RegularGridInterpolator((r[:,0],z[0,:]), logDens,\
                bounds_error=False, fill_value=-300.)

        points = numpy.empty((rr.size, 2))
        points[:,0] = rr.reshape((-1,))/AU
        points[:,1] = zz.reshape((-1,))/AU

        logDens_interp = f(points).reshape(rr.shape)

        Dens = numpy.exp(logDens_interp)

        return Dens

    def number_density(self, w1, w2, w3, gas=0, coordsys="spherical", \
            mstar=0.5):
        # Get the high resolution log of the gas density.

        r, z, logn = self.log_number_density_high(mstar=mstar, gas=gas)

        rho = numpy.sqrt(r**2 + z**2)
        theta = numpy.pi/2 - numpy.arctan(z / r)

        # Now, interpolate that density onto the actual grid of interest by
        # averaging all of the high resolution points that fall within the 
        # lower resolution cell.

        n = numpy.exp(logn)

        n_interp = scipy.stats.binned_statistic_2d(rho.flatten(), \
                theta.flatten(), n.flatten(), statistic="mean", \
                bins=(w1, w2)).statistic[:,:,numpy.newaxis]

        n_interp[numpy.isnan(n_interp)] = 0.

        return n_interp

    def temperature(self, x1, x2, x3, coordsys="spherical"):
        ##### Disk Parameters
        
        rin = self.rmin * AU
        rout = self.rmax * AU
        pltgas = self.pltgas
        tmid0 = self.tmid0
        tatm0 = self.tatm0
        zq0 = self.zq0 * AU
        delta = self.delta

        ##### Set up the coordinates

        if coordsys == "spherical":
            rt, tt, pp = numpy.meshgrid(x1*AU, x2, x3, indexing='ij')

            rr = rt*numpy.sin(tt)
            zz = rt*numpy.cos(tt)
        elif coordsys == "cartesian":
            xx, yy, zz = numpy.meshgrid(x1*AU, x2*AU, x3*AU, indexing='ij')

            rr = (xx**2 + yy**2)**0.5
        elif coordsys == "cylindrical":
            rr, pp, zz = numpy.meshgrid(x1*AU, x2, x3*AU, indexing='ij')

        ##### Make the dust density model for a protoplanetary disk.
        
        zq = zq0 * (rr / (1*AU))**1.3

        with numpy.errstate(divide="ignore"):
            tmid = numpy.where(rr > 0, tmid0 * (rr / (1*AU))**(-pltgas), 0.1)
            tatm = numpy.where(rr > 0, tatm0 * (rr / (1*AU))**(-pltgas), 0.1)

        t = numpy.zeros(tatm.shape)
        t[zz >= zq] = tatm[zz >= zq]
        t[zz < zq] = tatm[zz < zq] + (tmid[zz < zq] - tatm[zz < zq]) * \
                (numpy.cos(numpy.pi * zz[zz < zq] / (2*zq[zz < zq])))**(2*delta)

        ##### Catch an temperatures greater than 10000 K and set to that value,
        ##### as that is likely unphysical.

        t[t > 10000] = 10000
        
        return t

    def microturbulence(self, r, theta, phi):
        ##### Disk Parameters
        
        rin = self.rmin * AU
        rout = self.rmax * AU
        t0 = self.t0
        plt = self.plt

        ##### Set up the coordinates

        rt, tt, pp = numpy.meshgrid(r*AU, theta, phi,indexing='ij')

        rr = rt*numpy.sin(tt)
        zz = rt*numpy.cos(tt)

        # Calculate the sound speed.

        T = self.temperature(r, theta, phi, coordsys="spherical")

        # Parameterize in terms of the sound speed.
        
        mu = 2.37
        cs = (k*T/(mu*m_p))**0.5

        aturb = cs*self.aturb
        
        return aturb

    def velocity(self, r, theta, phi, mstar=0.5):
        rt, tt, pp = numpy.meshgrid(r*AU, theta, phi,indexing='ij')

        rr = rt*numpy.sin(tt)
        zz = rt*numpy.cos(tt)

        v_r = numpy.zeros(rr.shape)
        v_theta = numpy.zeros(rr.shape)

        # Calculate the Keplerian velocity.

        v_phi_kepler = numpy.sqrt(G*mstar*M_sun*rr**2/rt**3)

        # Calculate the pressure contribution to the velocity.

        R, z, logP = self.log_gas_pressure_high(mstar=mstar)
        R, z, logDens = self.log_gas_density_high(mstar=mstar)

        P = numpy.exp(logP)
        Dens = numpy.exp(logDens)
        Dens[logDens < -200.] = 0.

        dP = numpy.gradient(P, R[:,0]*AU, axis=0)

        with numpy.errstate(divide="ignore", invalid="ignore"):
            v_phi_pressure2 = numpy.where(Dens > 0, dP / Dens * R*AU, 0.)

        # Now, interpolate onto the proper grid.

        f = scipy.interpolate.RegularGridInterpolator((R[:,0],z[0,:]), \
                v_phi_pressure2, bounds_error=False, fill_value=0.)

        points = numpy.empty((rr.size, 2))
        points[:,0] = rr.reshape((-1,))/AU
        points[:,1] = zz.reshape((-1,))/AU

        v_phi_pressure2_interp = f(points).reshape(rr.shape)

        # Finally, add the two together.

        with numpy.errstate(divide="ignore", invalid="ignore"):
            v_phi = numpy.where(v_phi_kepler**2 + v_phi_pressure2_interp > 0, \
                    numpy.sqrt(v_phi_kepler**2 + v_phi_pressure2_interp), 1.0)

        return numpy.array((v_r, v_theta, v_phi))

